#file: pmix.pxi
#
# Copyright (c) 2022      Nanook Consulting. All rights reserved
#

from libc.string cimport memset, strncpy, strcpy, strlen, strdup
from libc.stdlib cimport malloc, realloc, free
from libc.string cimport memcpy
from cpython.mem cimport PyMem_Malloc, PyMem_Realloc, PyMem_Free
from cpython.pycapsule cimport PyCapsule_New, PyCapsule_GetPointer

# pull in all the constant definitions - we
# store them in a separate file for neatness
include "pmix_constants.pxi"

# provide a lock class for catching information
# returned in callback functions
class myLock(threading.Event):
    def __init__(self):
        threading.Event.__init__(self)
        self.event = threading.Event()
        self.status = PMIX_ERR_NOT_SUPPORTED
        self.sz = 0
        self.info = []

    def set(self, status):
        self.status = status
        self.event.set()

    def clear(self):
        self.event.clear()

    def wait(self):
        self.event.wait()

    def get_status(self):
        return self.status

    def cache_data(self, data, sz):
        self.data = array.array('B', data[0])
        # need to copy the data bytes as the
        # PMIx server will free it upon return
        n = 1
        while n < sz:
            self.data.append(data[n])
            n += 1
        self.sz = sz

    def fetch_data(self):
        return (self.data, self.sz)

    def cache_info(self, info:list):
        # need to copy the info array as the
        # PMIx server will free it upon execing
        # the callback function
        self.info = []
        for x in info:
            self.info.append(x)

    def fetch_info(self, info:list):
        for x in self.info:
            info.append(x)

ctypedef struct pmix_pyshift_t:
    char *op
    pmix_byte_object_t payload
    size_t idx
    pmix_modex_cbfunc_t modex
    pmix_status_t status
    pmix_byte_object_t bo
    pmix_byte_object_t *cred
    pmix_iof_channel_t channel
    pmix_nspace_t nspace
    pmix_proc_t source
    pmix_proc_t *proc
    pmix_pdata_t *pdata
    pmix_info_t *results
    size_t nresults
    pmix_info_t *info
    const char *data
    size_t ndata
    pmix_op_cbfunc_t op_cbfunc
    pmix_iof_cbfunc_t iof
    pmix_info_cbfunc_t query
    pmix_spawn_cbfunc_t spawn
    pmix_lookup_cbfunc_t lookup
    pmix_release_cbfunc_t release_fn
    pmix_event_notification_cbfunc_fn_t event_handler
    pmix_tool_connection_cbfunc_t toolconnected
    pmix_credential_cbfunc_t getcredential
    pmix_validation_cbfunc_t validationcredential
    pmix_info_cbfunc_t allocate
    pmix_info_cbfunc_t sessioncontrol
    void *notification_cbdata
    void *cbdata

cdef void iofhdlr_cache(capsule, ret):
    cdef pmix_pyshift_t *shifter
    cdef pmix_byte_object_t *bo
    shifter = <pmix_pyshift_t*>PyCapsule_GetPointer(capsule, "iofhdlr_cache")
    if NULL == shifter[0].payload.bytes:
        bo = NULL
    else:
        bo = &shifter[0].payload
    pyiofhandler(shifter[0].idx, shifter[0].channel, &shifter[0].source,
                 bo, shifter[0].info, shifter[0].ndata)
    if 0 < shifter[0].ndata:
        pmix_free_info(shifter[0].info, shifter[0].ndata)
    free(shifter[0].payload.bytes)
    return

cdef void event_cache_cb(capsule, ret):
    cdef pmix_pyshift_t *shifter
    shifter = <pmix_pyshift_t*>PyCapsule_GetPointer(capsule, "event_handler")
    pyeventhandler(shifter[0].idx, shifter[0].status, &shifter[0].source,
                   shifter[0].info, shifter[0].ndata,
                   shifter[0].results, shifter[0].nresults,
                   shifter[0].event_handler, shifter[0].notification_cbdata)

cdef void pmix_convert_locality(pmix_locality_t loc, pyloc:list):
    if PMIX_LOCALITY_NONLOCAL & loc:
        pyloc.append(PMIX_LOCALITY_NONLOCAL)
    if PMIX_LOCALITY_SHARE_HWTHREAD & loc:
        pyloc.append(PMIX_LOCALITY_SHARE_HWTHREAD)
    if PMIX_LOCALITY_SHARE_CORE & loc:
        pyloc.append(PMIX_LOCALITY_SHARE_CORE)
    if PMIX_LOCALITY_SHARE_L1CACHE & loc:
        pyloc.append(PMIX_LOCALITY_SHARE_L1CACHE)
    if PMIX_LOCALITY_SHARE_L2CACHE & loc:
        pyloc.append(PMIX_LOCALITY_SHARE_L2CACHE)
    if PMIX_LOCALITY_SHARE_L3CACHE & loc:
        pyloc.append(PMIX_LOCALITY_SHARE_L3CACHE)
    if PMIX_LOCALITY_SHARE_PACKAGE & loc:
        pyloc.append(PMIX_LOCALITY_SHARE_PACKAGE)
    if PMIX_LOCALITY_SHARE_NUMA & loc:
        pyloc.append(PMIX_LOCALITY_SHARE_NUMA)
    if PMIX_LOCALITY_SHARE_NODE & loc:
        pyloc.append(PMIX_LOCALITY_SHARE_NODE)
    if 0 == pyloc.len():
        pyloc.append(PMIX_LOCALITY_NONLOCAL)
    return

cdef void pmix_unload_argv(char **keys, argv:list):
    n = 0
    while NULL != keys[n]:
        mykey = keys[n].decode('ascii')
        argv.append(mykey)
        n += 1

cdef int pmix_load_argv(char **keys, argv:list):
    n = 0
    for a in argv:
        pya = a
        if isinstance(a, str):
            pya = a.encode('ascii')
        keys[n] = strdup(pya)
        n += 1
    keys[n] = NULL
    return PMIX_SUCCESS

cdef int pmix_load_darray(pmix_data_array_t *array, mytype, mylist:list):
    cdef pmix_info_t *infoptr;
    cdef pmix_value_t *valptr;
    mysize = len(mylist)
    if PMIX_INFO == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(pmix_info_t))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        infoptr = <pmix_info_t*>array[0].array
        rc = pmix_load_info(infoptr, mylist)
    elif PMIX_VALUE == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(pmix_value_t))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        valptr = <pmix_value_t*>array[0].array
        rc = pmix_load_value(valptr, mylist)
    elif PMIX_BOOL == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(int*))
        n = 0
        if not array[0].array:
            return PMIX_ERR_NOMEM
        boolptr = <int*>array[0].array
        for item in mylist:
            int_bool = pmix_bool_convert(item)
            if int_bool != 0 and int_bool != 1:
                return PMIX_ERR_BAD_PARAM
            boolptr[n] = int_bool
            n += 1
    elif PMIX_BYTE == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(uint8_t*))
        n = 0
        # byte val is uint8 type
        if not array[0].array:
            return PMIX_ERR_NOMEM
        bptr = <uint8_t*> array[0].array
        for item in mylist:
            if not isinstance(item, pmix_int_types):
                print("uint8 value declared but non-integer provided")
                return PMIX_ERR_TYPE_MISMATCH
            if item > 255:
                print("uint8 value is out of bounds")
                return PMIX_ERR_BAD_PARAM
            bptr[n] = int(item)
            n += 1
    elif PMIX_STRING == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(char*))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        strptr = <char**> array[0].array
        for item in mylist:
            if isinstance(item, str):
                pykey = item.encode('ascii')
            else:
                pykey = item
            try:
                strptr[n] = strdup(pykey)
            except:
                print("String value declared but non-string provided")
                return PMIX_ERR_TYPE_MISMATCH
            n += 1
    elif PMIX_SIZE == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(size_t))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        sptr = <size_t*> array[0].array
        for item in mylist:
            if not isinstance(item, pmix_int_types):
                print("size_t value declared but non-integer provided")
                return PMIX_ERR_TYPE_MISMATCH
            sptr[n] = int(item)
            n += 1
    elif PMIX_PID == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(pid_t))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        pidptr = <pid_t*> array[0].array
        for item in mylist:
            if not isinstance(item, pmix_int_types):
                print("pid_t value declared but non-integer provided")
                return PMIX_ERR_TYPE_MISMATCH
            pidptr[n] = int(item)
            n += 1
    elif PMIX_INT == mytype or PMIX_UINT == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(int))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        iptr = <int*> array[0].array
        for item in mylist:
            if not isinstance(item, pmix_int_types):
                print("int value declared but non-integer provided")
                return PMIX_ERR_TYPE_MISMATCH
            iptr[n] = int(item)
            n += 1
    elif PMIX_INT8 == mytype or PMIX_UINT8 == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(int8_t))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        i8ptr = <int8_t*> array[0].array
        for item in mylist:
            if not isinstance(item, pmix_int_types):
                print("8-bit int value declared but non-integer provided")
                return PMIX_ERR_TYPE_MISMATCH
            i8ptr[n] = int(item)
            n += 1
    elif PMIX_INT16 == mytype or PMIX_UINT16 == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(int16_t))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        i16ptr = <int16_t*> array[0].array
        for item in mylist:
            if not isinstance(item, pmix_int_types):
                print("16-bit int value declared but non-integer provided")
                return PMIX_ERR_TYPE_MISMATCH
            i16ptr[n] = int(item)
            n += 1
    elif PMIX_INT32 == mytype or PMIX_UINT32 == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(int32_t))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        i32ptr = <int32_t*> array[0].array
        for item in mylist:
            if not isinstance(item, pmix_int_types):
                print("32-bit int value declared but non-integer provided")
                return PMIX_ERR_TYPE_MISMATCH
            i32ptr[n] = int(item)
            n += 1
    elif PMIX_INT64 == mytype or PMIX_UINT64 == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(int64_t))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        i64ptr = <int64_t*> array[0].array
        for item in mylist:
            if not isinstance(item, pmix_int_types):
                print("64-bit int value declared but non-integer provided")
                return PMIX_ERR_TYPE_MISMATCH
            i64ptr[n] = int(item)
            n += 1
    elif PMIX_FLOAT == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(float))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        fptr = <float*> array[0].array
        for item in mylist:
            fptr[n] = float(item)
            n += 1
    elif PMIX_DOUBLE == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(double))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        dptr = <double*> array[0].array
        for item in mylist:
            dptr[n] = float(item)
            n += 1
            n += 1
    elif PMIX_TIMEVAL == mytype:
        # TODO: Not clear that "timeval" has the same size as
        # "struct timeval"
        array[0].array = PyMem_Malloc(mysize * sizeof(timeval))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        tvptr = <timeval*> array[0].array
        for item in mylist:
            tvptr[n].tv_sec  = item['sec']
            tvptr[n].tv_usec = item['usec']
            n += 1
    elif PMIX_TIME == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(time_t))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        tmptr = <time_t*> array[0].array
        for item in mylist:
            tmptr[n] = item
            n += 1
    elif PMIX_STATUS == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(int))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        stptr = <int*> array[0].array
        for item in mylist:
            stptr[n] = item
            n += 1
    elif PMIX_PROC_RANK == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(pmix_rank_t))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        rkptr = <pmix_rank_t*> array[0].array
        for item in mylist:
            rkptr[n] = item
            n += 1
    elif PMIX_PROC == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(pmix_proc_t))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        prcptr = <pmix_proc_t*> array[0].array
        for item in mylist:
            pmix_copy_nspace(prcptr[n].nspace, item['nspace'])
            prcptr[n].rank = item['rank']
            n += 1
    elif PMIX_BYTE_OBJECT == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(pmix_byte_object_t))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        boptr = <pmix_byte_object_t*>array[0].array
        for item in mylist:
            boptr[n].size = item['size']
            boptr[n].bytes = <char*> PyMem_Malloc(boptr[n].size)
            if not boptr[n].bytes:
                return PMIX_ERR_NOMEM
            pyarr = bytes(item['bytes'])
            pyptr = <const char*> pyarr
            memcpy(boptr[n].bytes, pyptr, boptr[n].size)
            n += 1
    elif PMIX_PERSISTENCE == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(pmix_persistence_t))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        perptr = <pmix_persistence_t*> array[0].array
        for item in mylist:
            perptr[n] = item
            n += 1
    elif PMIX_SCOPE == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(pmix_scope_t))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        scptr = <pmix_scope_t*> array[0].array
        for item in mylist:
            scptr[n] = item
            n += 1
    elif PMIX_RANGE == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(pmix_data_range_t))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        rgptr = <pmix_data_range_t*> array[0].array
        for item in mylist:
            rgptr[n] = item
            n += 1
    elif PMIX_PROC_STATE == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(pmix_proc_state_t))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        psptr = <pmix_proc_state_t*> array[0].array
        for item in mylist:
            psptr[n] = item
            n += 1
    elif PMIX_PROC_INFO == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(pmix_proc_info_t))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        piptr = <pmix_proc_info_t*> array[0].array
        for item in mylist:
            pmix_copy_nspace(piptr[n].proc.nspace, item['proc']['nspace'])
            piptr[n].proc.rank = item['proc']['rank']
            hostname = item['hostname']
            if isinstance(hostname, str):
                pyhostname = hostname.encode('ascii')
            else:
                pyhostname = hostname
            pyhostnameptr = <const char *>(pyhostname)
            piptr[n].hostname = strdup(pyhostnameptr)
            executable = item['executable']
            if isinstance(executable, str):
                pyexec = executable.encode('ascii')
            else:
                pyexec = executable
            pyexecptr = <const char *>(pyexec)
            piptr[n].executable_name = strdup(pyexecptr)
            piptr[n].pid = item['pid']
            piptr[n].exit_code = item['exitcode']
            piptr[n].state = item['state']
            n += 1
    elif PMIX_DATA_ARRAY == mytype:
        array[0].array = <pmix_data_array_t*> PyMem_Malloc(sizeof(pmix_data_array_t))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        daptr = <pmix_data_array_t*>array[0].array
        n = 0
        for item in mylist:
            daptr[n].type = item['val_type']
            daptr[n].size = len(item['value'])
            daptr[n].array = <pmix_data_array_t*> PyMem_Malloc(sizeof(pmix_data_array_t))
            if not daptr[n].array:
                return PMIX_ERR_NOMEM
            mydaptr = <pmix_data_array_t*>daptr[n].array
            try:
                return pmix_load_darray(mydaptr, daptr[n].type, item['value'])
            except:
                return PMIX_ERR_NOT_SUPPORTED
    elif PMIX_ALLOC_DIRECTIVE == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(pmix_alloc_directive_t))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        aldptr = <pmix_alloc_directive_t*> array[0].array
        for item in mylist:
            aldptr[n] = item
            n += 1
    elif PMIX_ENVAR == mytype:
        array[0].array = PyMem_Malloc(mysize * sizeof(pmix_envar_t))
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        envptr = <pmix_envar_t*> array[0].array
        for item in mylist:
            enval = item['envar']
            if isinstance(enval, str):
                pyns = enval.encode('ascii')
            else:
                pyns = enval
            pynsptr = <const char *>(pyns)
            envptr[n].envar = strdup(pynsptr)
            if isinstance(enval, str):
                pyns = enval.encode('ascii')
            else:
                pyns = enval
            pynsptr = <const char *>(pyns)
            envptr[n].value = strdup(pynsptr)
            pyseparator = <char>ord(item['separator'])
            envptr[n].separator = pyseparator
            n += 1
    else:
        print("UNRECOGNIZED DATA TYPE IN ARRAY "+str(array[0].type))
        return PMIX_ERR_NOT_SUPPORTED
    return PMIX_SUCCESS

cdef dict pmix_unload_darray(pmix_data_array_t *array):
    cdef pmix_info_t *infoptr;
    if PMIX_INFO == array.type:
        ilist = []
        n = 0
        infoptr = <pmix_info_t*>array[0].array
        rc = pmix_unload_info(infoptr, array.size, ilist)
        darray = {'type':array.type, 'array':ilist}
        return darray
    elif PMIX_BOOL == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        boolptr = <int*>array[0].array
        list = []
        n = 0
        while n < array.size:
            list.append(boolptr[n])
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_BYTE == array.type:
        # byte val is uint8 type
        if not array[0].array:
            return PMIX_ERR_NOMEM
        bptr = <uint8_t*> array[0].array
        list = []
        n = 0
        while n < array.size:
            list.append(bptr[n])
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_STRING == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        strptr = <char**> array[0].array
        strlist = []
        while n < array.size:
            pyb = strptr[n]
            pystr = pyb.decode("ascii")
            strlist.append(pystr)
            free(strptr[n])
            n += 1
        darray = {'type':array.type, 'array':strlist}
        return darray
    elif PMIX_SIZE == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        sptr = <size_t*> array[0].array
        list = []
        while n < array.size:
            list.append(sptr[n])
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_PID == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        pidptr = <pid_t*> array[0].array
        list = []
        while n < array.size:
            list.append(pidptr[n])
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_INT == array.type or PMIX_UINT == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        iptr = <int*> array[0].array
        list = []
        while n < array.size:
            list.append(iptr[n])
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_INT8 == array.type or PMIX_UINT8 == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        i8ptr = <int8_t*> array[0].array
        list = []
        while n < array.size:
            list.append(i8ptr[n])
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_INT16 == array.type or PMIX_UINT16 == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        i16ptr = <int16_t*> array[0].array
        list = []
        while n < array.size:
            list.append(i16ptr[n])
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_INT32 == array.type or PMIX_UINT32 == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        i32ptr = <int32_t*> array[0].array
        list = []
        while n < array.size:
            list.append(i32ptr[n])
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_INT64 == array.type or PMIX_UINT64 == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        i64ptr = <int64_t*> array[0].array
        list = []
        while n < array.size:
            list.append(i64ptr[n])
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_FLOAT == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        fptr = <float*> array[0].array
        list = []
        while n < array.size:
            list.append(fptr[n])
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_DOUBLE == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        dptr = <double*> array[0].array
        list = []
        while n < array.size:
            list.append(dptr[n])
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_TIMEVAL == array.type:
        # TODO: Not clear that "timeval" has the same size as
        # "struct timeval"
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        tvptr = <timeval*> array[0].array
        list = []
        while n < array.size:
            d = {'sec': tvptr[n].tv_sec, 'usec': tvptr[n].tv_usec}
            list.append(d)
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_TIME == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        tmptr = <time_t*> array[0].array
        list = []
        while n < array.size:
            list.append(tmptr[n])
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_STATUS == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        stptr = <int*> array[0].array
        list = []
        while n < array.size:
            list.append(stptr[n])
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_PROC_RANK == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        rkptr = <pmix_rank_t*> array[0].array
        list = []
        while n < array.size:
            list.append(rkptr[n])
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_PROC == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        prcptr = <pmix_proc_t*> array[0].array
        list = []
        while n < array.size:
            d = {'nspace': prcptr[n].nspace, 'rank': prcptr[n].rank}
            list.append(d)
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_BYTE_OBJECT == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        boptr = <pmix_byte_object_t*>array[0].array
        list = []
        while n < array.size:
            if not boptr[n].bytes:
                return PMIX_ERR_NOMEM
            d = {'bytes': boptr[n].bytes, 'size': boptr[n].size}
            list.append(d)
            PyMem_Free(boptr[n].bytes)
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_PERSISTENCE == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        perptr = <pmix_persistence_t*> array[0].array
        list = []
        while n < array.size:
            list.append(perptr[n])
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_SCOPE == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        scptr = <pmix_scope_t*> array[0].array
        list = []
        while n < array.size:
            list.append(scptr[n])
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_RANGE == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        rgptr = <pmix_data_range_t*> array[0].array
        list = []
        while n < array.size:
            list.append(rgptr[n])
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_PROC_STATE == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        psptr = <pmix_proc_state_t*> array[0].array
        list = []
        while n < array.size:
            list.append(psptr[n])
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_PROC_INFO == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        piptr = <pmix_proc_info_t*> array[0].array
        list = []
        while n < array.size:
            d = {'proc': {'nspace':piptr[n].proc.nspace,
            'rank':piptr[n].proc.rank}, 'hostname':piptr[n].hostname,
            'executable':piptr[n].executable_name, 'pid':piptr[n].pid,
            'exitcode':piptr[n].exit_code, 'state':piptr[n].state}
            list.append(d)
            free(piptr[n].hostname)
            free(piptr[n].executable_name)
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_DATA_ARRAY == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        daptr = <pmix_data_array_t*>array[0].array
        n = 0
        while n < array.size:
            if not daptr[n].array:
                return PMIX_ERR_NOMEM
            mydaptr = <pmix_data_array_t*>daptr[n].array
            try:
                rc = pmix_unload_darray(mydaptr)
                if rc != PMIX_SUCCESS:
                    return rc
            except:
                return PMIX_ERR_NOT_SUPPORTED
            n += 1
    elif PMIX_ALLOC_DIRECTIVE == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        aldptr = <pmix_alloc_directive_t*> array[0].array
        list = []
        while n < array.size:
            list.append(aldptr[n])
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_ENVAR == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        n = 0
        envptr = <pmix_envar_t*> array[0].array
        list = []
        while n < array.size:
            pyenv = envptr[n].envar
            pyval = envptr[n].value
            pysep = envptr[n].separator
            d = {'envar':pyenv, 'value':pyval, 'separator':pysep}
            list.append(d)
            free(envptr[n].value)
            free(envptr[n].envar)
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    elif PMIX_VALUE == array.type:
        if not array[0].array:
            return PMIX_ERR_NOMEM
        list = []
        n = 0
        valptr = <pmix_value_t*>array[0].array
        while n < array.size:
            d = pmix_unload_value(&valptr[n])
            list.append(d)
            n += 1
        darray = {'type':array.type, 'array':list}
        return darray
    else:
        print("UNRECOGNIZED DATA TYPE IN ARRAY: "+str(array[0].type))
        return PMIX_ERR_NOT_SUPPORTED
    return PMIX_SUCCESS

# provide conversion programs that translate incoming
# PMIx structures into Python dictionaries, and incoming
# arrays into Python lists of objects

def pmix_bool_convert(f):
    int_bool = PMIX_ERR_BAD_PARAM
    if isinstance(f, str):
        if f.startswith('t') or f.startswith('T'):
            int_bool = 1
        elif f.startswith('f') or f.startswith('F'):
            int_bool = 0
        else:
            print("Incorrect boolean value provided")
            int_bool = PMIX_ERR_BAD_PARAM
    elif isinstance(f, (int, float)):
        int_bool = f
    else:
        print("Unrecognized boolean value type provided")

    return int_bool

# Tuple of integer types.
#  Python 2.x distinguished int/long, but in
#  Python 3.x dropped long and just uses int or width specific variants.
pmix_int_types = (int,)

# provide a safe way to copy a Python nspace into
# the pmix_nspace_t structure that guarantees the
# array is NULL-terminated
cdef void pmix_copy_nspace(pmix_nspace_t nspace, ns):
    nslen = len(ns)
    if PMIX_MAX_NSLEN < nslen:
        nslen = PMIX_MAX_NSLEN
    if isinstance(ns, str):
        pyns = ns.encode('ascii')
    else:
        pyns = ns
    pynsptr = <const char *>(pyns)
    memset(nspace, 0, PMIX_MAX_NSLEN+1)
    memcpy(nspace, pynsptr, nslen)

# provide a safe way to copy a Python key into
# the pmix_key_t structure that guarantees the
# array is NULL-terminated
cdef void pmix_copy_key(pmix_key_t key, ky):
    klen = len(ky)
    if PMIX_MAX_KEYLEN < klen:
        klen = PMIX_MAX_KEYLEN
    if isinstance(ky, str):
        pykey = ky.encode('ascii')
    else:
        pykey = ky
    pykeyptr = <const char *>(pykey)
    memset(key, 0, PMIX_MAX_KEYLEN+1)
    if 'b' == ky[0]:
        memcpy(key, &pykeyptr[2], klen-3)
    else:
        memcpy(key, pykeyptr, klen)

# loads a python pmix regex into a python bytearray
cdef bytearray pmix_convert_regex(char *regex):
    if "pmix" == regex[:4].decode("ascii"):
        # remove null characters
        if b'\x00' in regex:
            regex.replace(b'\x00', '')
        ba = bytearray(regex)
    elif "blob" == regex[:4].decode("ascii"):
        sz_str    = len(regex)
        sz_prefix = 5
        # extract length of bytearray
        regex.split(b'\x00')
        len_bytearray = regex[1]
        length = len(len_bytearray) + sz_prefix + sz_str
        ba = bytearray(length)
        pyregex = <bytes> regex[:length]
        index = 0
        while index < length:
            ba[index] = pyregex[index]
            index += 1
    else:
        # last case with no ':' in string
        ba = bytearray(regex)
    return ba

# provide a function for transferring a Python 'value'
# object (a dict with value and val_type as keys)
# to a pmix_value_t
cdef int pmix_load_value(pmix_value_t *value, val:dict):
    if not isinstance(val['val_type'], pmix_int_types):
        return PMIX_ERR_BAD_PARAM
    value[0].type = val['val_type']

    if val['val_type'] == PMIX_BOOL:
        int_bool = pmix_bool_convert(val['value'])
        if int_bool != 0 and int_bool != 1:
            return PMIX_ERR_BAD_PARAM
        value[0].data.flag = int_bool

    elif val['val_type'] == PMIX_BYTE:
        # byte val is uint8 type
        if not isinstance(val['value'], pmix_int_types):
            print("uint8 value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value'] > 255:
            print("uint8 value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.byte = val['value']

    elif val['val_type'] == PMIX_STRING:
        if isinstance(val['value'], str):
            pykey = val['value'].encode('ascii')
        else:
            pykey = val['value']
        try:
            value[0].data.string = strdup(pykey)
        except:
            print("String value declared but non-string provided")
            return PMIX_ERR_TYPE_MISMATCH

    elif val['val_type'] == PMIX_SIZE:
        if not isinstance(val['value'], pmix_int_types):
            print("size_t value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        value[0].data.size = val['value']

    elif val['val_type'] == PMIX_PID:
        if not isinstance(val['value'], pmix_int_types):
            print("pid value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value'] < 0:
            print("pid value is negative")
            return PMIX_ERR_BAD_PARAM
        value[0].data.pid = val['value']

    elif val['val_type'] == PMIX_INT:
        if not isinstance(val['value'], pmix_int_types):
            print("integer value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        value[0].data.integer = val['value']

    elif val['val_type'] == PMIX_INT8:
        if not isinstance(val['value'], pmix_int_types):
            print("int8 value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value'] > 127 or val['value'] < -128:
            print("int8 value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.int8 = val['value']

    elif val['val_type'] == PMIX_INT16:
        if not isinstance(val['value'], pmix_int_types):
            print("int16 value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value'] > 32767 or val['value'] < -32768:
            print("int16 value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.int16 = val['value']

    elif val['val_type'] == PMIX_INT32:
        if not isinstance(val['value'], pmix_int_types):
            print("int32 value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value'] > 2147483647 or val['value'] < -2147483648:
            print("int32 value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.int32 = val['value']

    elif val['val_type'] == PMIX_INT64:
        if not isinstance(val['value'], pmix_int_types):
            print("int64 value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value'] > (2147483647*2147483647) or val['value'] < -(2147483648*2147483648):
            print("int64 value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.int64 = val['value']

    elif val['val_type'] == PMIX_UINT:
        if not isinstance(val['value'], pmix_int_types):
            print("integer value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value'] < 0:
            print("uint value out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.uint = val['value']

    elif val['val_type'] == PMIX_UINT8:
        if not isinstance(val['value'], pmix_int_types):
            print("uint8 value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value'] > 255 or val['value'] < 0:
            print("uint8 value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.uint8 = val['value']

    elif val['val_type'] == PMIX_UINT16:
        if not isinstance(val['value'], pmix_int_types):
            print("uint16 value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value'] > 65536 or val['value'] < 0:
            print("uint16 value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.uint16 = val['value']

    elif val['val_type'] == PMIX_UINT32:
        if not isinstance(val['value'], pmix_int_types):
            print("uint32 value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value'] > (65536*65536) or val['value'] < 0:
            print("uint32 value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.uint32 = val['value']

    elif val['val_type'] == PMIX_UINT64:
        if not isinstance(val['value'], pmix_int_types):
            print("int64 value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value'] > (2147483648*2147483648):
            print("uint64 value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.uint64 = val['value']

    elif val['val_type'] == PMIX_FLOAT:
        float_val = float(val['value'])
        if not isinstance(float_val, float):
            return PMIX_ERR_TYPE_MISMATCH
        value[0].data.fval = float_val

    elif val['val_type'] == PMIX_DOUBLE:
        double_val = float(val['value'])
        if not isinstance(double_val, float):
            return PMIX_ERR_TYPE_MISMATCH
        value[0].data.dval = double_val

    # TODO: need a way to verify usable timevals passed in?
    elif val['val_type'] == PMIX_TIMEVAL:
        value[0].data.tv.tv_sec  = val['value']['sec']
        value[0].data.tv.tv_usec = val['value']['usec']

    elif val['val_type'] == PMIX_TIME:
        value[0].data.time = val['val_type']

    elif val['val_type'] == PMIX_STATUS:
        if not isinstance(val['value'], int):
            return PMIX_ERR_TYPE_MISMATCH
        value[0].data.status = val['value']

    elif val['val_type'] == PMIX_PROC_RANK:
        if not isinstance(val['value'], int):
            return PMIX_ERR_TYPE_MISMATCH
        if val['value'] > (65536*65536) or val['value'] < 0:
            print("uint32 value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.rank = val['value']

    elif val['val_type'] == PMIX_PROC_NSPACE:
        value[0].data.nspace = <pmix_nspace_t*> PyMem_Malloc(sizeof(pmix_nspace_t))
        if not value[0].data.nspace:
            return PMIX_ERR_NOMEM
        pmix_copy_nspace(value[0].data.nspace[0], val['value'])

    elif val['val_type'] == PMIX_PROC:
        value[0].data.proc = <pmix_proc_t*> PyMem_Malloc(sizeof(pmix_proc_t))
        if not value[0].data.proc:
            return PMIX_ERR_NOMEM
        # TODO: check nspace val is a char here
        pmix_copy_nspace(value[0].data.proc[0].nspace, val['value']['nspace'])
        # pmix_rank_t is defined as uint32
        if not isinstance(val['value']['rank'], pmix_int_types):
            print("uint32 value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value']['rank'] > (65536*65536):
            print("uint32 value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.proc[0].rank = val['value']['rank']

    # TODO: pmix byte object conversion isn't working
    elif val['val_type'] == PMIX_BYTE_OBJECT:
        value[0].data.bo.size = val['value']['size']
        value[0].data.bo.bytes = <char*> PyMem_Malloc(value[0].data.bo.size)
        if not value[0].data.bo.bytes:
            return PMIX_ERR_NOMEM
        pyptr = <char*>val['value']['bytes']
        memcpy(value[0].data.bo.bytes, pyptr, value[0].data.bo.size)

    # regex is can be a string or a byte object
    elif val['val_type'] == PMIX_REGEX:
        regex = val['value']
        ba = pmix_convert_regex(regex)
        if not isinstance(ba, bytearray):
            return PMIX_ERR_TYPE_MISMATCH
        value[0].data.bo.size  = len(val['value'])
        value[0].data.bo.bytes = <char*> PyMem_Malloc(value[0].data.bo.size)
        if not value[0].data.bo.bytes:
            return PMIX_ERR_NOMEM
        pyptr = <char*>ba
        memcpy(value[0].data.bo.bytes, pyptr, value[0].data.bo.size)

    elif val['val_type'] == PMIX_PERSISTENCE:
        # pmix_persistence_t is defined as uint8
        if not isinstance(val['value'], pmix_int_types):
            print("uint8 value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value'] > 255 or val['value'] < 0:
            print("uint8 value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.persist = val['value']

    elif val['val_type'] == PMIX_SCOPE:
        # pmix_scope_t is defined as uint8
        if not isinstance(val['value'], pmix_int_types):
            print("uint8 value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value'] > 255:
            print("uint8 value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.scope = val['value']

    elif val['val_type'] == PMIX_RANGE:
        # pmix_data_range_t is defined as uint8
        if not isinstance(val['value'], pmix_int_types):
            print("uint8 value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value'] > 255:
            print("uint8 value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.range = val['value']

    elif val['val_type'] == PMIX_PROC_STATE:
        # pmix_proc_state_t is defined as uint8
        if not isinstance(val['value'], pmix_int_types):
            print("uint8 value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value'] > 255:
            print("uint8 value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.state = val['value']

    elif val['val_type'] == PMIX_PROC_INFO:
        value[0].data.pinfo = <pmix_proc_info_t*> PyMem_Malloc(sizeof(pmix_proc_info_t))
        if not value[0].data.pinfo:
            return PMIX_ERR_NOMEM
        # TODO: verify nspace is copied correctly
        pmix_copy_nspace(value[0].data.pinfo[0].proc.nspace, val['value']['proc']['nspace'])
        if not isinstance(val['value']['proc']['rank'], pmix_int_types):
            print("uint32 value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value']['proc']['rank'] > (65536*65536):
            print("uint32 value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.pinfo[0].proc.rank = val['value']['proc']['rank']
        hostname = val['value']['hostname']
        if isinstance(hostname, str):
            pyhostname = hostname.encode('ascii')
        else:
            pyhostname = hostname
        pyhostnameptr = <const char *>(pyhostname)
        value[0].data.pinfo[0].hostname = strdup(pyhostnameptr)
        executable = val['value']['executable']
        if isinstance(executable, str):
            pyexec = executable.encode('ascii')
        else:
            pyexec = executable
        pyexecptr = <const char *>(pyexec)
        value[0].data.pinfo[0].executable_name = strdup(pyexecptr)
        # TODO: check this is a pid type
        value[0].data.pinfo[0].pid = val['value']['pid']
        if not isinstance(val['value']['exitcode'], int):
            print("value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        value[0].data.pinfo[0].exit_code = val['value']['exitcode']
        if not isinstance(val['value']['state'], pmix_int_types):
            print("uint8 value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value']['state'] > 255:
            print("uint8 value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.pinfo[0].state = val['value']['state']

    elif val['val_type'] == PMIX_DATA_ARRAY:
        value[0].data.darray = <pmix_data_array_t*> PyMem_Malloc(sizeof(pmix_data_array_t))
        if not value[0].data.darray:
            return PMIX_ERR_NOMEM
        value[0].data.darray[0].type = val['value']['type']
        value[0].data.darray[0].size = len(val['value']['array'])
        try:
            # assume pmix_load_darray does own type checks
            # it should return with an error code inside that
            # function if there is one
            pmix_load_darray(value[0].data.darray, value[0].data.darray[0].type, val['value']['array'])
        except:
            return PMIX_ERR_NOT_SUPPORTED

    elif val['val_type'] == PMIX_ALLOC_DIRECTIVE:
        if not isinstance(val['value'], pmix_int_types):
            print("allocdirective value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value'] > 255:
            print("allocdirective value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.adir = val['value']

    elif val['val_type'] == PMIX_RESBLOCK_DIRECTIVE:
        if not isinstance(val['value'], pmix_int_types):
            print("resblockdirective value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value'] > 255:
            print("allocdirective value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.rbdir = val['value']

    elif val['val_type'] == PMIX_ENVAR:
        enval = val['value']['envar']
        if isinstance(enval, str):
            pyns = enval.encode('ascii')
        else:
            pyns = enval
        pynsptr = <const char *>(pyns)
        value[0].data.envar.envar = strdup(pynsptr)
        enval = val['value']['value']
        if isinstance(enval, str):
            pyns = enval.encode('ascii')
        else:
            pyns = enval
        pynsptr = <const char *>(pyns)
        value[0].data.envar.value = strdup(pynsptr)
        # TODO: way/function to verify char type
        enval = val['value']['separator']
        if isinstance(enval, pmix_int_types):
            value[0].data.envar.separator = enval
        else:
            value[0].data.envar.separator = ord(enval)

    elif val['val_type'] == PMIX_COORD:
        value[0].data.coord = <pmix_coord_t*> PyMem_Malloc(sizeof(pmix_coord_t))
        if not value[0].data.coord:
            return PMIX_ERR_NOMEM
        value[0].data.coord[0].view = val['value']['view']
        dims = val['value']['dims']
        value[0].data.coord[0].coord = <uint32_t*> PyMem_Malloc(dims * sizeof(uint32_t))
        if not value[0].data.coord[0].coord:
            return PMIX_ERR_NOMEM
        pyptr = <char*> val['value']['coord']
        memcpy(value[0].data.coord[0].coord, pyptr, dims * sizeof(uint32_t))
        value[0].data.coord[0].dims = dims

    elif val['val_type'] == PMIX_LINK_STATE:
        if not isinstance(val['value'], pmix_int_types):
            print("linkstate value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value'] > 255:
            print("linkstate value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.linkstate = val['value']

    elif val['val_type'] == PMIX_JOB_STATE:
        if not isinstance(val['value'], pmix_int_types):
            print("jobstate value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value'] > 255:
            print("linkstate value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.jstate = val['value']

    elif val['val_type'] == PMIX_TOPO:
        return PMIX_ERR_NOT_SUPPORTED

    elif val['val_type'] == PMIX_PROC_CPUSET:
        return PMIX_ERR_NOT_SUPPORTED

    elif val['val_type'] == PMIX_LOCTYPE:
        if not isinstance(val['value'], pmix_int_types):
            print("locality value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value'] > 65535:
            print("locality value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.locality = val['value']

    elif val['val_type'] == PMIX_GEOMETRY:
        value[0].data.geometry = <pmix_geometry_t*> PyMem_Malloc(sizeof(pmix_geometry_t))
        if not value[0].data.geometry:
            return PMIX_ERR_NOMEM
        value[0].data.geometry[0].fabric = val['value']['fabric']
        uuid = val['value']['uuid']
        if isinstance(uuid, str):
            pyuuid = uuid.encode('ascii')
        else:
            pyuuid = uuid
        pyuuidptr = <const char *>(pyuuid)
        value[0].data.geometry[0].uuid = strdup(pyuuidptr)
        osname = val['value']['osname']
        if isinstance(osname, str):
            pyosname = osname.encode('ascii')
        else:
            pyosname = osname
        pyosnameptr = <const char *>(pyosname)
        value[0].data.geometry[0].osname = strdup(pyosnameptr)
        ncoords = val['value']['ncoords']
        value[0].data.geometry[0].coordinates = <pmix_coord_t*> PyMem_Malloc(ncoords * sizeof(pmix_coord_t))
        if not value[0].data.geometry[0].coordinates:
            return PMIX_ERR_NOMEM
        n = 0
        for k in val['value']['coords']:
            value[0].data.geometry[0].coordinates[n].view = k['view']
            dims = k['dims']
            value[0].data.geometry[0].coordinates[n].coord = <uint32_t*> PyMem_Malloc(dims * sizeof(uint32_t))
            if not value[0].data.geometry[0].coordinates[n].coord:
                return PMIX_ERR_NOMEM
            pyptr = <char*> k['coord']
            memcpy(value[0].data.geometry[0].coordinates[n].coord, pyptr, dims * sizeof(uint32_t))
            value[0].data.geometry[0].coordinates[n].dims = dims
            n = n + 1
        return PMIX_ERR_NOT_SUPPORTED

    elif val['val_type'] == PMIX_DEVTYPE:
        if not isinstance(val['value'], pmix_int_types):
            print("devtype value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value'] > 18446744073709551615:
            print("devtype value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.devtype = val['value']

    elif val['val_type'] == PMIX_DEVICE:
        value[0].data.device = <pmix_device_t*> PyMem_Malloc(sizeof(pmix_device_t))
        if not value[0].data.device:
            return PMIX_ERR_NOMEM
        uuid = val['value']['uuid']
        if isinstance(uuid, str):
            pyuuid = uuid.encode('ascii')
        else:
            pyuuid = uuid
        pyuuidptr = <const char *>(pyuuid)
        value[0].data.device[0].uuid = strdup(pyuuidptr)
        osname = val['value']['osname']
        if isinstance(osname, str):
            pyosname = osname.encode('ascii')
        else:
            pyosname = osname
        pyosnameptr = <const char *>(pyosname)
        value[0].data.device[0].osname = strdup(pyosnameptr)
        if not isinstance(val['value']['type'], pmix_int_types):
            print("devtype value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value']['type'] > 18446744073709551615:
            print("devtype value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.device[0].type = val['value']['type']

    elif val['val_type'] == PMIX_DEVICE_DIST:
        value[0].data.devdist = <pmix_device_distance_t*> PyMem_Malloc(sizeof(pmix_device_distance_t))
        if not value[0].data.devdist:
            return PMIX_ERR_NOMEM
        uuid = val['value']['uuid']
        if isinstance(uuid, str):
            pyuuid = uuid.encode('ascii')
        else:
            pyuuid = uuid
        pyuuidptr = <const char *>(pyuuid)
        value[0].data.devdist[0].uuid = strdup(pyuuidptr)
        osname = val['value']['osname']
        if isinstance(osname, str):
            pyosname = osname.encode('ascii')
            pyosname = osname
        pyosnameptr = <const char *>(pyosname)
        value[0].data.devdist[0].osname = strdup(pyosnameptr)
        if not isinstance(val['value']['mindist'], pmix_int_types):
            print("mindist value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value']['mindist'] > 65535:
            print("mindist value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.devdist[0].mindist = val['value']['mindist']
        if not isinstance(val['value']['maxdist'], pmix_int_types):
            print("maxdist value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value']['maxdist'] > 65535:
            print("maxdist value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.devdist[0].maxdist = val['value']['maxdist']

    elif val['val_type'] == PMIX_ENDPOINT:
        value[0].data.endpoint = <pmix_endpoint_t*> PyMem_Malloc(sizeof(pmix_endpoint_t))
        if not value[0].data.endpoint:
            return PMIX_ERR_NOMEM
        uuid = val['value']['uuid']
        if isinstance(uuid, str):
            pyuuid = uuid.encode('ascii')
        else:
            pyuuid = uuid
        pyuuidptr = <const char *>(pyuuid)
        value[0].data.endpoint[0].uuid = strdup(pyuuidptr)
        osname = val['value']['osname']
        if isinstance(osname, str):
            pyosname = osname.encode('ascii')
            pyosname = osname
        pyosnameptr = <const char *>(pyosname)
        value[0].data.endpoint[0].osname = strdup(pyosnameptr)
        value[0].data.endpoint[0].endpt.size = val['value']['endpt']['size']
        value[0].data.endpoint[0].endpt.bytes = <char*> PyMem_Malloc(value[0].data.endpoint[0].endpt.size)
        if not value[0].data.endpoint[0].endpt.bytes:
            return PMIX_ERR_NOMEM
        pyptr = <char*>val['value']['endpt']['bytes']
        memcpy(value[0].data.endpoint[0].endpt.bytes, pyptr, value[0].data.endpoint[0].endpt.size)

    elif val['val_type'] == PMIX_DATA_BUFFER:
        return PMIX_ERR_NOT_SUPPORTED

    elif val['val_type'] == PMIX_RESOURCE_UNIT:
        value[0].data.resunit = <pmix_resource_unit_t*> PyMem_Malloc(sizeof(pmix_resource_unit_t))
        if not value[0].data.resunit:
            return PMIX_ERR_NOMEM
        if not isinstance(val['value']['type'], pmix_int_types):
            print("devtype value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value']['type'] > 18446744073709551615:
            print("devtype value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.resunit[0].type = val['value']['type']
        if not isinstance(val['value']['count'], pmix_int_types):
            print("count value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value']['count'] > 18446744073709551615:
            print("count value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.resunit[0].count = val['value']['count']

    elif val['val_type'] == PMIX_NODE_PID:
        value[0].data.nodepid = <pmix_node_pid_t*> PyMem_Malloc(sizeof(pmix_node_pid_t))
        if not value[0].data.nodepid:
            return PMIX_ERR_NOMEM
        hostname = val['value']['hostname']
        if isinstance(hostname, str):
            pyhostname = hostname.encode('ascii')
        else:
            pyhostname = hostname
        pyhostnameptr = <const char *>(pyhostname)
        value[0].data.nodepid[0].hostname = strdup(pyhostnameptr)
        if not isinstance(val['value']['nodeid'], pmix_int_types):
            print("nodeid value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value']['nodeid'] > 4294967295:
            print("nodeid value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.nodepid[0].nodeid = val['value']['nodeid']
        if not isinstance(val['value']['pid'], pmix_int_types):
            print("pid value declared but non-integer provided")
            return PMIX_ERR_TYPE_MISMATCH
        if val['value']['pid'] > 4294967295:
            print("pid value is out of bounds")
            return PMIX_ERR_BAD_PARAM
        value[0].data.nodepid[0].pid = val['value']['pid']

    else:
        print("UNRECOGNIZED VALUE TYPE: "+str(val['val_type']))
        return PMIX_ERR_NOT_SUPPORTED
    return PMIX_SUCCESS

cdef dict pmix_unload_value(const pmix_value_t *value):

    if PMIX_BOOL == value[0].type:
        if value[0].data.flag:
            return {'value':True, 'val_type':PMIX_BOOL}
        else:
            return {'value':False, 'val_type':PMIX_BOOL}

    elif PMIX_BYTE == value[0].type:
        return {'value':value[0].data.byte, 'val_type':PMIX_BYTE}

    elif PMIX_STRING == value[0].type:
        pyb = value[0].data.string
        try:
            pystr = pyb.decode("ascii")
        except:
            pystr = pyb
        return {'value':pystr, 'val_type':PMIX_STRING}

    elif PMIX_SIZE == value[0].type:
        return {'value':value[0].data.size, 'val_type':PMIX_SIZE}

    elif PMIX_PID == value[0].type:
        return {'value':value[0].data.pid, 'val_type':PMIX_PID}

    elif PMIX_INT == value[0].type:
        return {'value':value[0].data.integer, 'val_type':PMIX_INT}

    elif PMIX_INT8 == value[0].type:
        return {'value':value[0].data.int8, 'val_type':PMIX_INT8}

    elif PMIX_INT16 == value[0].type:
        return {'value':value[0].data.int16, 'val_type':PMIX_INT16}

    elif PMIX_INT32 == value[0].type:
        return {'value':value[0].data.int32, 'val_type':PMIX_INT32}

    elif PMIX_INT64 == value[0].type:
        return {'value':value[0].data.int64, 'val_type':PMIX_INT64}

    elif PMIX_UINT == value[0].type:
        return {'value':value[0].data.uint, 'val_type':PMIX_UINT}

    elif PMIX_UINT8 == value[0].type:
        return {'value':value[0].data.uint8, 'val_type':PMIX_UINT8}

    elif PMIX_UINT16 == value[0].type:
        return {'value':value[0].data.uint16, 'val_type':PMIX_UINT16}

    elif PMIX_UINT32 == value[0].type:
        return {'value':value[0].data.uint32, 'val_type':PMIX_UINT32}

    elif PMIX_UINT64 == value[0].type:
        return {'value':value[0].data.uint64, 'val_type':PMIX_UINT64}

    elif PMIX_FLOAT == value[0].type:
        return {'value':value[0].data.fval, 'val_type':PMIX_FLOAT}

    elif PMIX_DOUBLE == value[0].type:
        return {'value':value[0].data.dval, 'val_type':PMIX_DOUBLE}

    elif PMIX_TIMEVAL == value[0].type:
        return {'value':{'sec':value[0].data.tv.tv_sec, 'usec':value[0].data.tv.tv_usec},
        'val_type':PMIX_TIMEVAL}

    elif PMIX_TIME == value[0].type:
        return {'value':value[0].data.time, 'val_type':PMIX_TIME}

    elif PMIX_STATUS == value[0].type:
        return {'value':value[0].data.status, 'val_type':PMIX_STATUS}

    elif PMIX_PROC_RANK == value[0].type:
        return {'value':value[0].data.rank, 'val_type':PMIX_PROC_RANK}

    elif PMIX_PROC_NSPACE == value[0].type:
        pyns = (<bytes>value[0].data.nspace[0]).decode('UTF-8')
        return {'value':pyns, 'val_type':PMIX_PROC_NSPACE}

    elif PMIX_PROC == value[0].type:
        pyns = (<bytes>value[0].data.proc[0].nspace).decode('UTF-8')
        return {'value':{'nspace':pyns, 'rank':value[0].data.proc[0].rank}, 'val_type':PMIX_PROC}

    elif PMIX_BYTE_OBJECT == value[0].type:
        mybytes = <bytes>value[0].data.bo.bytes[value[0].data.bo.size]
        return {'value':{'bytes':mybytes, 'size':value[0].data.bo.size}, 'val_type':PMIX_BYTE_OBJECT}

    elif PMIX_REGEX == value[0].type:
        return {'value': value[0].data.bo.bytes, 'val_type': PMIX_REGEX}

    elif PMIX_PERSISTENCE == value[0].type:
        return {'value':value[0].data.persist, 'val_type':PMIX_PERSISTENCE}

    elif PMIX_SCOPE == value[0].type:
        return {'value':value[0].data.scope, 'val_type':PMIX_SCOPE}

    elif PMIX_RANGE == value[0].type:
        return {'value':value[0].data.range, 'val_type':PMIX_RANGE}

    elif PMIX_PROC_STATE == value[0].type:
        return {'value':value[0].data.state, 'val_type':PMIX_PROC_STATE}

    elif PMIX_PROC_INFO == value[0].type:
        pins = (<bytes>value[0].data.pinfo[0].proc.nspace).decode('UTF-8')
        pirk = value[0].data.pinfo[0].proc.rank
        pihost = (<bytes>value[0].data.pinfo[0].hostname).decode('UTF-8')
        pexec = (<bytes>value[0].data.pinfo[0].executable_name).decode('UTF-8')
        pipid = value[0].data.pinfo[0].pid
        piex = value[0].data.pinfo[0].exit_code
        pist = value[0].data.pinfo[0].state
        pians = {'proc': {'nspace':pins, 'rank':pirk}, 'hostname': pihost, 'executable': pexec, 'pid': pipid, 'exitcode': piex, 'state': pist}
        return {'value':pians, 'val_type':PMIX_PROC_INFO}

    elif PMIX_DATA_ARRAY == value[0].type:
        try:
            # assume pmix_unload_darray does own type checks
            # it should return with an error code inside that
            # function if there is one
            darray = pmix_unload_darray(value[0].data.darray)
            return {'value':darray, 'val_type':PMIX_DATA_ARRAY}
        except:
            return PMIX_ERR_NOT_SUPPORTED

    elif PMIX_ALLOC_DIRECTIVE == value[0].type:
        return {'value':value[0].data.adir, 'val_type':PMIX_ALLOC_DIRECTIVE}

    elif PMIX_RESBLOCK_DIRECTIVE == value[0].type:
        return {'value':value[0].data.rbdir, 'val_type':PMIX_RESBLOCK_DIRECTIVE}

    elif PMIX_ENVAR == value[0].type:
        pyenv = (<bytes>value[0].data.envar.envar).decode('UTF-8')
        pyval = (<bytes>value[0].data.envar.value).decode('UTF-8')
        pysep = value[0].data.envar.separator
        pyenvans = {'envar': pyenv, 'value': pyval, 'separator': pysep}
        return {'value':pyenvans, 'val_type':PMIX_ENVAR}

    elif PMIX_COORD == value[0].type:
        pyview = value[0].data.coord[0].view
        pydims = value[0].data.coord[0].dims
        mybytes = <bytes>value[0].data.coord[0].coord[pydims]
        return {'value': {'view': pyview, 'coord': mybytes, 'dims': pydims}, 'val_type': PMIX_COORD}

    elif PMIX_LINK_STATE == value[0].type:
        return {'value': value[0].data.linkstate, 'val_type': PMIX_LINK_STATE}

    elif PMIX_JOB_STATE == value[0].type:
        return {'value': value[0].data.jstate, 'val_type': PMIX_JOB_STATE}

    elif PMIX_TOPO == value[0].type:
        return {'value': None, 'val_type': PMIX_TOPO}

    elif PMIX_PROC_CPUSET == value[0].type:
        return {'value': None, 'val_type': PMIX_PROC_CPUSET}

    elif PMIX_LOCTYPE == value[0].type:
        return {'value': value[0].data.locality, 'val_type': PMIX_LOCTYPE}

    elif PMIX_GEOMETRY == value[0].type:
        pyfabric = value[0].data.geometry[0].fabric
        pyuuid = (<bytes>value[0].data.geometry[0].uuid).decode('UTF-8')
        pyosname = (<bytes>value[0].data.geometry[0].osname).decode('UTF-8')
        pyncoord = value[0].data.geometry[0].ncoords
        pylist = []
        for n in range(pyncoord):
            pyview = value[0].data.geometry.coordinates[n].view
            pydims = value[0].data.geometry.coordinates[n].dims
            mybytes = <bytes>value[0].data.geometry.coordinates[n].coord[pydims]
            pyc = {'view': pyview, 'coord': mybytes, 'dims': pydims}
            pylist.append(pyc)
        return {'value': {'fabric': pyfabric, 'uuid': pyuuid, 'osname': pyosname, 'coords': pylist}, 'val_type': PMIX_GEOMETRY}

    elif PMIX_DEVTYPE == value[0].type:
        return {'value': value[0].data.devtype, 'val_type': PMIX_DEVTYPE}

    elif PMIX_DEVICE == value[0].type:
        pyuuid = (<bytes>value[0].data.device[0].uuid).decode('UTF-8')
        pyosname = (<bytes>value[0].data.device[0].osname).decode('UTF-8')
        pyval = {'uuid': pyuuid, 'osname': pyosname, 'type': value[0].data.device[0].type}
        return {'value': pyval, 'val_type': PMIX_DEVICE}

    elif PMIX_DEVICE_DIST == value[0].type:
        pyuuid = (<bytes>value[0].data.devdist[0].uuid).decode('UTF-8')
        pyosname = (<bytes>value[0].data.devdist[0].osname).decode('UTF-8')
        pyval = {'uuid': pyuuid, 'osname': pyosname, 'mindist': value[0].data.devdist[0].mindist, 'maxdist': value[0].data.devdist[0].maxdist}
        return {'value': pyval, 'val_type': PMIX_DEVICE_DIST}

    elif PMIX_ENDPOINT == value[0].type:
        pyuuid = (<bytes>value[0].data.endpoint[0].uuid).decode('UTF-8')
        pyosname = (<bytes>value[0].data.endpoint[0].osname).decode('UTF-8')
        pysize = value[0].data.endpoint[0].size
        mybytes = <bytes>value[0].data.endpoint[0].bytes[pysize]
        pyval = {'uuid': pyuuid, 'osname': pyosname, 'endpt': {'bytes': mybytes, 'size': pysize}}
        return {'value': pyval, 'val_type': PMIX_ENDPOINT}

    elif PMIX_RESOURCE_UNIT == value[0].type:
        return {'value': {'type': value[0].data.resunit[0].type, 'count': value[0].data.resunit[0].count}, 'val_type': PMIX_RESOURCE_UNIT}

    elif PMIX_NODE_PID == value[0].type:
        pyhostname = (<bytes>value[0].data.nodepid[0].hostname).decode('UTF-8')
        return {'value': {'hostname': pyhostname, 'nodeid': value[0].data.nodepid[0].nodeid, 'pid': value[0].data.nodepid[0].pid}, 'val_type': PMIX_NODE_PID}

    else:
        print("Unload_value: provided type is unknown", value[0].type)
        return {'value': None, 'val_type': PMIX_UNDEF}

cdef void pmix_destruct_value(pmix_value_t *value):
    if value[0].type == PMIX_STRING:
        free(value[0].data.string);

cdef void pmix_free_value(self, pmix_value_t *value):
    pmix_destruct_value(value);
    PyMem_Free(value)

# Convert a dictionary of key-value pairs into an
# array of pmix_info_t structs
#
# @array [INPUT]
#        - malloc'd array of pmix_info_t structs
#
# @dicts [INPUT]
#          - a list of dictionaries, where each
#            dictionary has a key, value, and val_type
#            defined as such:
#            [{key:y, value:val, val_type:ty}, … ]
#
cdef int pmix_load_info(pmix_info_t *array, dicts:list):
    n = 0
    for d in dicts:
        pykey = str(d['key'])
        pmix_copy_key(array[n].key, pykey)
        try:
            array[n].flags = d['flags']
        except:
            pass
        val = {'value':d['value'], 'val_type':d['val_type']}
        rc = pmix_load_value(&array[n].value, val)
        if PMIX_SUCCESS != rc:
            return rc
        n += 1
    return PMIX_SUCCESS

# Allocate memory and load pmix info structs
#
# @array [INPUT]
#          - array of pmix_info_t structs
#
# @ninfo [INPUT]
#          - length of the list of dictionaries
#
# @dicts [INPUT]
#          - a list of dictionaries, where each
#            dictionary has a key, value, and val_type
#            defined as such:
#            [{key:y, value:val, val_type:ty}, … ]
#
cdef int pmix_alloc_info(pmix_info_t **info_ptr, size_t *ninfo, dicts:list):
    # Convert any provided dictionary to an array of pmix_info_t
    if dicts is not None:
        ninfo[0] = len(dicts)
        if 0 < ninfo[0]:
            info_ptr[0] = <pmix_info_t*>malloc(ninfo[0] * sizeof(pmix_info_t))
            if not info_ptr[0]:
                return PMIX_ERR_NOMEM
            rc = pmix_load_info(info_ptr[0], dicts)
            if PMIX_SUCCESS != rc:
                pmix_free_info(info_ptr[0], ninfo[0])
                return rc
        else:
            info_ptr[0] = NULL
            ninfo[0] = 0
    else:
        info_ptr[0] = NULL
        ninfo[0] = 0
    return PMIX_SUCCESS

cdef int pmix_unload_info(const pmix_info_t *info, size_t ninfo, ilist:list):
    cdef char* kystr
    cdef size_t n
    n = 0
    while n < ninfo:
        kystr = strdup(info[n].key)
        # pmix_unload_value returns a python dict of val, val_type
        val = pmix_unload_value(&info[n].value)
        if val['val_type'] == PMIX_UNDEF:
            return PMIX_ERR_NOT_SUPPORTED
        d     = {}
        pykey = str(kystr.decode("ascii"))
        free(kystr)
        d['key']      = pykey
        # TODO: don't know how to decide if flags was defined or not??
        d['flags']    = info[n].flags
        d['value']    = val['value']
        d['val_type'] = val['val_type']
        ilist.append(d)
        n += 1
    return PMIX_SUCCESS

cdef void pmix_destruct_info(pmix_info_t *info):
    pmix_destruct_value(&info[0].value)

# Free a malloc'd array of pmix_info_t structures
#
# @array [INPUT]
#        - array of pmix_info_t to be free'd
#
# @sz [INPUT]
#     - number of elements in array
cdef void pmix_free_info(pmix_info_t *array, size_t sz):
    n = 0
    while n < sz:
        pmix_destruct_info(&array[n])
        n += 1
    free(array)


# Convert array of resource units into a python list of
# dictionaries
cdef int pmix_unload_units(const pmix_resource_unit_t *units, size_t nunits, ilist:list):
    cdef size_t n
    n = 0
    while n < nunits:
        d     = {}
        d['type']      = units[n].type
        d['count']    = units[n].count
        ilist.append(d)
        n += 1
    return PMIX_SUCCESS


# Convert list of python pmix_pdata_t dicts into an
# array of pmix_pdata_t structs
#
# @array [INPUT]
#        - malloc'd array of pmix_pdata_t structs
#
# @pdata [INPUT]
#          - a list of dictionaries, where each
#            dictionary has a key, value, val_type,
#            and proc keys
cdef int pmix_load_pdata(pmix_pdata_t *array, data:list):
    n = 0
    for d in data:
        pykey = str(d['key'])
        pmix_copy_key(array[n].key, pykey)
        val = {'value':d['value'], 'val_type':d['val_type']}
        rc = pmix_load_value(&array[n].value, val)
        array[n].proc.rank = d['proc']['rank']
        pmix_copy_nspace(array[n].proc.nspace, d['proc']['nspace'])
        if PMIX_SUCCESS != rc:
            return rc
        n += 1
    return PMIX_SUCCESS

cdef int pmix_unload_pdata(const pmix_pdata_t *pdata, size_t npdata, ilist:list):
    cdef char* kystr
    cdef size_t n = 0
    while n < npdata:
        val = pmix_unload_value(&pdata[n].value)
        if val['val_type'] == PMIX_UNDEF:
            return PMIX_ERR_NOT_SUPPORTED
        d     = {}
        kystr = strdup(pdata[n].key)
        pykey = kystr.decode("ascii")
        free(kystr)
        d['key']      = pykey
        kystr = strdup(pdata[n].proc.nspace)
        myns = kystr.decode('ascii')
        free(kystr)
        proc = {'nspace':myns, 'rank': pdata[n].proc.rank}
        d['proc']     = proc
        d['value']    = val['value']
        d['val_type'] = val['val_type']
        ilist.append(d)
        n += 1
    return PMIX_SUCCESS

cdef void pmix_destruct_pdata(pmix_pdata_t *pdata):
    pmix_destruct_value(&pdata[0].value)

# Free a malloc'd array of pmix_pdata_t structures
#
# @array [INPUT]
#        - array of pmix_pdata_t to be free'd
#
# @sz [INPUT]
#     - number of elements in array
cdef void pmix_free_pdata(pmix_pdata_t *array, size_t sz):
    n = 0
    while n < sz:
        pmix_destruct_pdata(&array[n])
        n += 1
    PyMem_Free(array)

cdef int pmix_unload_queries(const pmix_query_t *queries, size_t nqueries, ilist:list):
    cdef char* kystr
    cdef size_t n = 0
    keylist = []
    qualist = []
    query = {}
    while n < nqueries:
        rc = pmix_unload_argv(queries[n].keys, keylist)
        pmix_unload_info(queries[n].qualifiers, queries[n].nqual, qualist)
        query['keys']       = keylist
        query['qualifiers'] = qualist
        ilist.append(query)
        n += 1
    return PMIX_SUCCESS


# Free a malloc'd array of pmix_query_t structs to free
#
# @array [INPUT]
#        - pmix_query_t queries to be free'd
#
# @sz [INPUT]
#     - number of elements in array
cdef void pmix_free_queries(pmix_query_t *queries, size_t sz):
    n = 0
    while n < sz:
        if queries[n].keys != NULL:
            j = 0
            while NULL != queries[n].keys[j]:
                PyMem_Free(queries[n].keys[j])
                j += 1
            PyMem_Free(queries[n].keys)
        if queries[n].qualifiers != NULL:
            pmix_free_info(queries[n].qualifiers, queries[n].nqual)
        n += 1
    if queries != NULL:
        PyMem_Free(queries)

# Convert a list of (nspace, rank) tuples into an
# array of pmix_proc_t structs
#
# @proc [INPUT]
#       - malloc'd array of pmix_proc_t structs
#
# @peers [INPUT]
#       - list of (nspace,rank) tuples
#
cdef int pmix_load_procs(pmix_proc_t *proc, peers:list):
    n = 0
    for p in peers:
        pmix_copy_nspace(proc[n].nspace, p['nspace'])
        proc[n].rank = p['rank']
        n += 1
    return PMIX_SUCCESS

cdef int pmix_unload_procs(const pmix_proc_t *procs, size_t nprocs, peers:list):
    cdef char* kystr
    n = 0
    while n < nprocs:
        kystr = strdup(procs[n].nspace)
        myns = kystr.decode('ascii')
        free(kystr)
        peers.append({'nspace':myns, 'rank':procs[n].rank})
        n += 1
    return PMIX_SUCCESS

# Free a malloc'd array of pmix_proc_t structures
#
# @array [INPUT]
#        - array of pmix_proc_t to be free'd
#
# @sz [INPUT]
#     - number of elements in array
#
cdef void pmix_free_procs(pmix_proc_t *array, size_t sz):
    PyMem_Free(array)

cdef void pmix_unload_bytes(char *data, size_t ndata, blist:list):
    cdef size_t n = 0
    while n < ndata:
        blist.append(data[n])
        n += 1

cdef void pmix_free_apps(pmix_app_t *array, size_t sz):
    n = 0
    while n < sz:
        free(array[n].cmd)
        # need to free the argv and env arrays
        free(array[n].cwd)
        if 0 < array[n].ninfo:
            pmix_free_info(array[n].info, array[n].ninfo)
        n += 1

cdef void pmix_unload_apps(const pmix_app_t *apps, size_t napps, pyapps:list):
    cdef size_t n = 0
    while n < napps:
        myapp = {}
        myapp['cmd'] = str(apps[n].cmd)
        myargv = []
        if NULL != apps[n].argv:
            pmix_unload_argv(apps[n].argv, myargv)
            myapp['argv'] = myargv
        myenv = []
        if NULL != apps[n].env:
            pmix_unload_argv(apps[n].env, myenv)
            myapp['env'] = myenv
        myapp['maxprocs'] = apps[n].maxprocs
        keyvals = {}
        if NULL != apps[n].info:
            pmix_unload_info(apps[n].info, apps[n].ninfo, keyvals)
            myapp['info'] = keyvals
        pyapps.append(myapp)
        n += 1

cdef int pmix_load_apps(pmix_app_t *apps, pyapps:list):
    cdef size_t m
    cdef size_t n
    cdef char** argv
    n = 0
    for p in pyapps:
        pycmd = str(p['cmd']).encode('ascii')
        try:
            apps[n].cmd = strdup(pycmd)
        except:
            return PMIX_ERR_TYPE_MISMATCH

        try:
            if p['argv'] is not None and 0 < len(p['argv']):
                m = len(p['argv']) + 1
            else:
                m = 2
            argv = <char**> malloc(m * sizeof(char*))
            if not argv:
                return PMIX_ERR_NOMEM
            memset(argv, 0, m)
            if p['argv'] is not None and 0 < len(p['argv']):
                pmix_load_argv(argv, p['argv'])
            else:
                pmix_load_argv(argv, [p['cmd']])
        except:
            argv = <char**> malloc(2 * sizeof(char*))
            memset(argv, 0, 2)
            rc = pmix_load_argv(argv, [p['cmd']])
            if PMIX_SUCCESS != rc:
                free(argv)
                return rc
        apps[n].argv = argv
        apps[n].env = NULL
        try:
            if p['env'] is not None:
                m = len(p['env']) + 1
                env = <char**> malloc(m * sizeof(char*))
                if not env:
                    return PMIX_ERR_NOMEM
                memset(env, 0, m)
                pmix_load_argv(env, p['env'])
                apps[n].env = env
        except:
            pass

        try:
            pycwd = str(p['cwd']).encode('ascii')
            apps[n].cwd = strdup(pycwd)
        except:
            pycwd = os.getcwd()
            pycwd = pycwd.encode('ascii')
            apps[n].cwd = strdup(pycwd)

        apps[n].info = NULL
        apps[n].ninfo = 0
        try:
            if p['info'] is not None:
                apps[n].ninfo = len(p['info'])
                apps[n].info =  <pmix_info_t*> malloc(apps[n].ninfo * sizeof(pmix_info_t))
                if not apps[n].info:
                    return PMIX_ERR_NOMEM
                rc = pmix_load_info(apps[n].info, p['info'])
                if PMIX_SUCCESS != rc:
                    return rc
        except:
            pass

        apps[n].maxprocs = 1
        try:
            apps[n].maxprocs = p['maxprocs']
        except:
            pass

        n += 1
    return PMIX_SUCCESS



#!/usr/bin/env python3

import time, sys
from pmix import *

test_count = 0
test_fails = 0
test_complete = 0

def test_put(client, scope, ky, val):
    print("PUT")
    global test_count, test_fails
    test_count = test_count + 1
    rc = client.put(PMIX_GLOBAL, "mykey", {'value':1, 'val_type':PMIX_INT32})
    if rc != 0 and rc != -47:
        test_fails = test_fails + 1
        print("PUT TEST FAILED: ", client.error_string(rc))

def test_commit(client):
    print("COMMIT")
    global test_count, test_fails
    test_count = test_count + 1
    rc = client.commit()
    if rc != 0 and rc != -47:
        test_fails = test_fails + 1
        print("COMMIT TEST FAILED: ", client.error_string(rc))

def test_fence(client, procs, info):
    print("FENCE")
    global test_count, test_fails
    test_count = test_count + 1
    rc = client.fence(procs, info)
    # don't count 'not supported' as a test fail
    if rc != 0 and rc != -47:
        test_fails = test_fails + 1
        print("FENCE TEST FAILED: ", client.error_string(rc))
    if rc == -47:
        print("FENCE call " + client.error_string(rc))

def test_get(client, proc, ky, dicts):
    print("GET")
    global test_count, test_fails
    test_count = test_count + 1
    rc, get_val = client.get(proc, ky, dicts)
    if rc != 0 and rc != -47:
        test_fails = test_fails + 1
        print("GET TEST FAILED: ", client.error_string(rc))
    print("GET_VAL RETURNED: ", get_val)

def test_publish(client, dicts):
    print("PUBLISH")
    global test_count, test_fails
    test_count = test_count + 1
    rc = client.publish(dicts)
    if rc != 0 and rc != -47:
        test_fails = test_fails + 1
        print("PUBLISH TEST FAILED: ", client.error_string(rc))

def test_lookup(client, data, dicts):
    print("LOOKUP")
    global test_count, test_fails
    test_count = test_count + 1
    rc, pdata = client.lookup(data, dicts)
    if rc != 0 and rc != -47:
        test_fails = test_fails + 1
        print("LOOKUP TEST FAILED: ", client.error_string(rc))
    print("PDATA RETURNED: ", pdata)

def test_unpublish(client, pykeys, dicts):
    print("UNPUBLISH")
    global test_count, test_fails
    test_count = test_count + 1
    rc = client.unpublish(pykeys, dicts)
    if rc != 0 and rc != -47:
        test_fails = test_fails + 1
        print("UNPUBLISH TEST FAILED: ", client.error_string(rc))

def test_query(client, pyqueries):
    print("QUERY")
    global test_count, test_fails
    test_count = test_count + 1
    rc, pyresults = client.query(pyqueries)
    if rc != 0 and rc != -47:
        test_fails = test_fails + 1
        print("QUERY TEST FAILED: ", client.error_string(rc))
    print("QUERY INFO RETURNED: ", pyresults)

def test_pyhandler(evid:int, st:int, pysource:dict, pyinfo:list, pyresults:list):
    global test_complete
    print("PYHANDLER", file=sys.stderr)
    status = PMIX_EVENT_ACTION_COMPLETE
    results = [{'key':'eventkey', 'value':'testevent', 'val_type':PMIX_STRING}]
    print("PYHANDLER RETURNED: ", pyinfo)
    test_complete = 1
    return status, results

def test_register_event_handler(client, pycodes, info, test_pyhandler):
    print("REGISTER EVENT HANDLER", file=sys.stderr)
    global test_count, test_fails
    test_count = test_count + 1
    rc, hdlr = client.register_event_handler(pycodes, info, test_pyhandler)
    if rc != 0 and rc != -47:
        test_fails = test_fails + 1
        print("REGISTER EVENT HANDLER TEST FAILED: ", client.error_string(rc))

def main():
    global test_complete
    foo = PMIxClient()
    print("Testing PMIx ", foo.get_version())
    info = [{'key':PMIX_PROGRAMMING_MODEL, 'flags': 0, 'value':'TEST', 'val_type':PMIX_STRING},
            {'key':PMIX_MODEL_LIBRARY_NAME, 'flags': 0, 'value':'PMIX', 'val_type':PMIX_STRING}]
    my_result,myname = foo.init(info)
    print("Init result ", my_result, file=sys.stderr)
    if 0 != my_result:
        print("FAILED TO INIT")
        exit(1)

    # put value into keystore
    test_put(foo, PMIX_GLOBAL, "mykey", {'value':1, 'val_type':PMIX_INT32})
    test_put(foo, PMIX_GLOBAL, "secondkey", {'value':2, 'val_type':PMIX_INT32})

    # commit it
    test_commit(foo)

    # execute fence
    procs = []
    info = []
    test_fence(foo, procs, info)

    info = []
    test_get(foo, {'nspace':"testnspace", 'rank': 0}, "mykey", info)

    # Get all keys for this proc
    test_get(foo, {'nspace':"testnspace", 'rank': 0}, None, info)

    # test a fence that should return not_supported because
    # we pass a required attribute that doesn't exist
    procs = []
    info = [{'key': 'ARBITRARY', 'flags': PMIX_INFO_REQD, 'value':10, 'val_type':PMIX_INT}]
    rc = test_fence(foo, procs, info)

    info = [{'key': 'ARBITRARY', 'flags': None, 'value':10, 'val_type':PMIX_INT}]
    test_publish(foo, info)

    test_fence(foo, procs, info)

    pdata_key = [{'key':'ARBITRARY'}]
    info = []
    test_lookup(foo, pdata_key, info)

    pykeys = ['ARBITRARY']
    info = []
    test_unpublish(foo, pykeys, info)

    test_fence(foo, procs, info)

    pyq = [{'keys':[PMIX_QUERY_PSET_NAMES], 'qualifiers':[]}]
    test_query(foo, pyq)

    pycodes = []
    code = PMIX_MODEL_DECLARED
    pycodes.append(code)
    info = [{'key': PMIX_EVENT_HDLR_NAME, 'value': 'SIMPCLIENT-MODEL', 'val_type': PMIX_STRING}]
    test_register_event_handler(foo, pycodes, info, test_pyhandler)

    loopcount = 0
    while test_complete != 1 and loopcount < 3:
        time.sleep(1)
        loopcount = loopcount + 1

    if test_complete == 0:
        print("MODEL EVENT FAILED TO ARRIVE", file=sys.stderr)

    # finalize
    info = []
    foo.finalize(info)
    print("Client finalize complete")

    # count client wrapper tests that returned with no error
    total_passed = test_count - test_fails
    print(str(total_passed) + " / " + str(test_count) + " TESTS PASSED")
if __name__ == '__main__':
    main()



#!/usr/bin/env python3

from pmix import *
from os import *
import time

def main():
    foo = PMIxTool()
    print("Testing PMIx Tool ", foo.get_version())
    # construct a session tmpdir
    uid = os.geteuid()
    gid = os.getegid()
    basename = os.path.basename(__file__)
    tmpdir = os.getenv('TMPDIR')
    sessiondir = os.path.join(tmpdir, basename + '.session.' + str(uid) + '.' + str(gid))
    info = [{'key':PMIX_LAUNCHER, 'value':True, 'val_type':PMIX_BOOL},
            {'key':PMIX_SERVER_TOOL_SUPPORT, 'value':True, 'val_type':PMIX_BOOL},
            {'key':PMIX_SERVER_TMPDIR, 'value':sessiondir, 'val_type':PMIX_STRING}]
    (my_result, myproc) = foo.init(info)
    if 0 != my_result:
        print("FAILED TO INIT", foo.error_string(my_result))
        exit(1)

    # construct an application to spawn
    app = {'cmd':'hostname', 'maxprocs':1}
    foo.spawn(None, [app])
    time.sleep(2.0)
    # finalize
    foo.finalize()
    print("Tool finalize complete")
if __name__ == '__main__':
    main()



#!/usr/bin/env python3

from pmix import *
import signal, time
import asyncio
from concurrent.futures import ThreadPoolExecutor
global killer

# Our event loop
event_loop = None

# where local data is published for testing
pmix_locdata = []

def start_event_loop():
    global event_loop
    if None == event_loop:
        event_loop = asyncio.new_event_loop()
        ThreadPoolExecutor().submit(event_loop.run_forever)

def stop_event_loop():
    global event_loop
    event_loop.call_soon_threadsafe(event_loop.stop)
    while event_loop.is_running():
        time.sleep(0.01)
    event_loop.close()

class GracefulKiller:
  kill_now = False
  def __init__(self):
    signal.signal(signal.SIGINT, self.exit_gracefully)
    signal.signal(signal.SIGTERM, self.exit_gracefully)

  def exit_gracefully(self,signum, frame):
    self.kill_now = True

def _clientconnected(proc:tuple is not None, cbfunc:pypmix_op_cbfunc_t, cbdata:dict):
    print("CLIENT CONNECTED SHIFTED", proc)
    cbfunc(PMIX_SUCCESS, cbdata)

def clientconnected(proc:tuple is not None, cbfunc:pypmix_op_cbfunc_t, cbdata:dict):
    print("CLIENT CONNECTED", proc)
    event_loop.call_soon_threadsafe(_clientconnected, proc, cbfunc, cbdata)
    return PMIX_SUCCESS

def _clientfinalized(proc:tuple is not None, cbfunc:pypmix_op_cbfunc_t, cbdata:dict):
    print("CLIENT FINALIZED SHIFTED", proc)
    cbfunc(PMIX_SUCCESS, cbdata)

def clientfinalized(proc:tuple is not None, cbfunc:pypmix_op_cbfunc_t, cbdata:dict):
    print("CLIENT FINALIZED", proc)
    event_loop.call_soon_threadsafe(_clientfinalized, proc, cbfunc, cbdata)
    return PMIX_SUCCESS

def _clientfence(args:dict is not None, cbfunc:pypmix_modex_cbfunc_t, cbdata:dict):
    # check directives
    print("CLIENTFENCE SHIFTED")
    output = bytearray(0)
    try:
        if args['directives'] is not None:
            for d in args['directives']:
                # these are each an info dict
                if "pmix" not in d['key']:
                    # we do not support such directives - see if
                    # it is required
                    try:
                        if d['flags'] & PMIX_INFO_REQD:
                            # return an error
                            return PMIX_ERR_NOT_SUPPORTED
                    except:
                        #it can be ignored
                        pass
    except:
        pass
    print("COMPLETE")
    cbfunc(PMIX_SUCCESS, output, cbdata)

def clientfence(args:dict is not None, cbfunc:pypmix_modex_cbfunc_t, cbdata:dict):
    # check directives
    print("CLIENTFENCE")
    event_loop.call_soon_threadsafe(_clientfence, args, cbfunc, cbdata)
    return PMIX_SUCCESS

def _clientpublish(args:dict is not None, cbfunc:pypmix_op_cbfunc_t, cbdata:dict):
    print("SERVER: PUBLISH SHIFTED")
    for d in args['directives']:
        pdata = {}
        pdata['proc'] = args['proc']
        pdata['key']            = d['key']
        pdata['value']          = d['value']
        pdata['val_type']       = d['val_type']
        pmix_locdata.append(pdata)
    cbfunc(PMIX_SUCCESS, cbdata)


def clientpublish(args:dict is not None, cbfunc:pypmix_op_cbfunc_t, cbdata:dict):
    print("SERVER: PUBLISH")
    event_loop.call_soon_threadsafe(_clientpublish, args, cbfunc, cbdata)
    return PMIX_SUCCESS

def _clientunpublish(args:dict is not None, cbfunc:pypmix_op_cbfunc_t, cbdata:dict):
    print("SERVER: UNPUBLISH SHIFTED")
    for k in args['keys']:
        for d in pmix_locdata:
            if k == d['key']:
                pmix_locdata.remove(d)
    cbfunc(PMIX_SUCCESS, cbdata)

def clientunpublish(args:dict is not None, cbfunc:pypmix_op_cbfunc_t, cbdata:dict):
    print("SERVER: UNPUBLISH")
    event_loop.call_soon_threadsafe(_clientunpublish, args, cbfunc, cbdata)
    return PMIX_SUCCESS

def _clientlookup(args:dict is not None, cbfunc:pypmix_lookup_cbfunc_t, cbdata:dict):
    print("SERVER: LOOKUP SHIFTED")
    ret_pdata = []
    for k in args['keys']:
        for d in pmix_locdata:
            if k.decode('ascii') == d['key']:
                ret_pdata.append(d)
    # return rc and pdata
    cbfunc(PMIX_SUCCESS, ret_pdata, cbdata)

def clientlookup(args:dict is not None, cbfunc:pypmix_lookup_cbfunc_t, cbdata:dict):
    print("SERVER: LOOKUP")
    event_loop.call_soon_threadsafe(_clientlookup, args, cbfunc, cbdata)
    return PMIX_SUCCESS

def _clientquery(args:dict is not None, cbfunc:pypmix_info_cbfunc_t, cbdata:dict):
    print("SERVER: QUERY SHIFTED")
    # return a python info list of dictionaries
    info = {}
    results = []
    rc = PMIX_ERR_NOT_FOUND
    # find key we passed in to client, and
    # if it matches return fake PSET_NAME
    # since RM actually assigns this, we
    # just return arbitrary name if key is
    # found
    find_str = 'pmix.qry.psets'
    for q in args['queries']:
        for k in q['keys']:
            if k == find_str:
                info = {'key': find_str, 'value': 'PSET_NAME', 'val_type': PMIX_STRING}
                results.append(info)
                rc = PMIX_SUCCESS
    cbfunc(rc, results, cbdata)

def clientquery(args:dict is not None, cbfunc:pypmix_info_cbfunc_t, cbdata:dict):
    print("SERVER: QUERY")
    event_loop.call_soon_threadsafe(_clientquery, args, cbfunc, cbdata)
    return PMIX_SUCCESS

def _client_register_events(args:dict is not None, cbfunc:pypmix_op_cbfunc_t, cbdata:dict):
    print("CLIENT REGISTER EVENTS SHIFTED", args['codes'])
    cbfunc(PMIX_SUCCESS, cbdata)


def client_register_events(args:dict is not None, cbfunc:pypmix_op_cbfunc_t, cbdata:dict):
    print("CLIENT REGISTER EVENTS", args['codes'])
    event_loop.call_soon_threadsafe(_client_register_events, args, cbfunc, cbdata)
    return PMIX_SUCCESS



#!/usr/bin/env python3

from pmix import *
import signal, time
import os
import select
import subprocess

global killer

class GracefulKiller:
  kill_now = False
  def __init__(self):
    signal.signal(signal.SIGINT, self.exit_gracefully)
    signal.signal(signal.SIGTERM, self.exit_gracefully)

  def exit_gracefully(self,signum, frame):
    self.kill_now = True

def clientconnected(proc:tuple is not None):
    print("CLIENT CONNECTED", proc)
    return PMIX_SUCCESS

def clientfinalized(proc:tuple is not None):
    print("CLIENT FINALIZED", proc)
    return PMIX_SUCCESS

def clientfence(args:dict is not None):
    print("SERVER FENCE", args)
    return PMIX_SUCCESS

def main():
    try:
        foo = PMIxServer()
    except:
        print("FAILED TO CREATE SERVER")
        exit(1)
    print("Testing server version ", foo.get_version())
    args = [{'key':PMIX_SERVER_SCHEDULER, 'value':'T', 'val_type':PMIX_BOOL}]
    map = {'clientconnected': clientconnected,
           'clientfinalized': clientfinalized,
           'fencenb': clientfence}
    my_result = foo.init(args, map)
    print("Testing PMIx_Initialized")
    rc = foo.initialized()
    print("Initialized: ", rc)
    vers = foo.get_version()
    print("Version: ", vers)

    # Register a fabric
    rc,fab = foo.fabric_register(None)
    print("Fabric registered: ", foo.error_string(rc))

    # setup the application
    (rc, regex) = foo.generate_regex(["test000","test001","test002"])
    print("Node regex, rc: ", regex, foo.error_string(rc))
    (rc, ppn) = foo.generate_ppn(["0,1,2", "3,4,5", "6,7"])
    print("PPN, rc: ", ppn, foo.error_string(rc))
    darray = {'type':PMIX_INFO, 'array':[{'key':PMIX_ALLOC_FABRIC_ID,
                            'value':'SIMPSCHED.net', 'val_type':PMIX_STRING},
                           {'key':PMIX_ALLOC_FABRIC_SEC_KEY, 'value':'T',
                            'val_type':PMIX_BOOL},
                           {'key':PMIX_SETUP_APP_ENVARS, 'value':'T',
                            'val_type':PMIX_BOOL}]}
    kyvals = [{'key':PMIX_NODE_MAP, 'value':regex, 'val_type':PMIX_STRING},
              {'key':PMIX_PROC_MAP, 'value':ppn, 'val_type':PMIX_STRING},
              {'key':PMIX_ALLOC_FABRIC, 'value':darray, 'val_type':PMIX_DATA_ARRAY}]

    appinfo = []
    rc, appinfo = foo.setup_application("SIMPSCHED", kyvals)
    print("SETUPAPP: ", appinfo)

    rc = foo.setup_local_support("SIMPSCHED", appinfo)
    print("SETUPLOCAL: ", foo.error_string(rc))

    # get our environment as a base
    env = os.environ.copy()

    # register an nspace for the client app
    kvals = [{'key':PMIX_NODE_MAP, 'value':regex, 'val_type':PMIX_STRING},
             {'key':PMIX_PROC_MAP, 'value':ppn, 'val_type':PMIX_STRING},
             {'key':PMIX_UNIV_SIZE, 'value':1, 'val_type':PMIX_UINT32},
             {'key':PMIX_JOB_SIZE, 'value':1, 'val_type':PMIX_UINT32}]
    print("REGISTERING NSPACE")
    rc = foo.register_nspace("testnspace", 1, kvals)
    print("RegNspace ", foo.error_string(rc))

    # register a client
    uid = os.getuid()
    gid = os.getgid()
    print("REGISTERING CLIENT")
    rc = foo.register_client({'nspace':"testnspace", 'rank':0}, uid, gid)
    print("RegClient ", foo.error_string(rc))

    # setup the fork
    rc = foo.setup_fork({'nspace':"testnspace", 'rank':0}, env)
    print("SetupFrk", foo.error_string(rc))

    # setup the client argv
    args = ["./client.py"]
    # open a subprocess with stdout and stderr
    # as distinct pipes so we can capture their
    # output as the process runs
    p = subprocess.Popen(args, env=env,
        stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    # define storage to catch the output
    stdout = []
    stderr = []
    # loop until the pipes close
    while True:
        reads = [p.stdout.fileno(), p.stderr.fileno()]
        ret = select.select(reads, [], [])

        stdout_done = True
        stderr_done = True

        for fd in ret[0]:
            # if the data
            if fd == p.stdout.fileno():
                read = p.stdout.readline()
                if read:
                    read = read.decode('utf-8').rstrip()
                    print('stdout: ' + read)
                    stdout_done = False
            elif fd == p.stderr.fileno():
                read = p.stderr.readline()
                if read:
                    read = read.decode('utf-8').rstrip()
                    print('stderr: ' + read)
                    stderr_done = False

        if stdout_done and stderr_done:
            break

    print("FINALIZING")
    foo.finalize()

if __name__ == '__main__':
    global killer
    killer = GracefulKiller()
    main()



#!/usr/bin/env python3

from pmix import *
import time
import threading

termEvent = threading.Event()

def default_evhandler(evhdlr:int, status:int,
                      source:dict, info:list, results:list):
    print("DEFAULT HANDLER")
    return PMIX_EVENT_ACTION_COMPLETE,None

def model_evhandler(evhdlr:int, status:int,
                    source:dict, info:list, results:list):
    global termEvent

    print("MODEL HANDLER")
    termEvent.set()
    return PMIX_EVENT_ACTION_COMPLETE,None

def main():
    global termEvent
    
    foo = PMIxClient()
    print("Testing PMIx ", foo.get_version())
    info = [{'key':PMIX_PROGRAMMING_MODEL, 'value':'TEST', 'val_type':PMIX_STRING},
            {'key':PMIX_MODEL_LIBRARY_NAME, 'value':'PMIX', 'val_type':PMIX_STRING}]
    my_result = foo.init(info)
    print("Init result ", my_result)
    if 0 != my_result[0]:
        print("FAILED TO INIT")
        exit(1)
    # register an event handler
    rc,myevhndlr = foo.register_event_handler(None, None, default_evhandler)
    print("REGISTER DEFAULT", foo.error_string(rc))
    # register the model handler
    rc,mymodelhndlr = foo.register_event_handler([PMIX_MODEL_DECLARED], None, model_evhandler)
    print("REGISTER MODEL", foo.error_string(rc))

    # try putting something
    print("PUT")
    rc = foo.put(PMIX_GLOBAL, "mykey", {'value':1, 'val_type':PMIX_INT32})
    print("Put result ",foo.error_string(rc));
    # commit it
    print("COMMIT")
    rc = foo.commit()
    print ("Commit result ", foo.error_string(rc))
    # execute fence
    print("FENCE")
    procs = []
    info = []
    rc = foo.fence(procs, info)
    print("Fence result ", foo.error_string(rc))
    print("GET")
    info = []
    rc, get_val = foo.get({'nspace':"testnspace", 'rank': 0}, "mykey", info)
    print("Get result: ", foo.error_string(rc))
    print("Get value returned: ", get_val)
    # test a fence that should return not_supported because
    # we pass a required attribute that doesn't exist
    procs = []
    info = [{'key': 'ARBITRARY', 'flags': PMIX_INFO_REQD, 'value':10, 'val_type':PMIX_INT}]
    rc = foo.fence(procs, info)
    print("Fence should be not supported:", foo.error_string(rc))
    # wait for model event
    while not termEvent.is_set():
        time.sleep(0.001)
    # finalize
    info = []
    foo.finalize(info)
    print("Client finalize complete")
if __name__ == '__main__':
    main()



#!/usr/bin/env python3
#
# Copyright (c) 2022      Nanook Consulting. All rights reserved

import os, os.path, sys, shutil, signal
from optparse import OptionParser, OptionGroup

takeconst = False
takeapis = False
takedtypes = False
takemacros = False

def signal_handler(signal, frame):
    print("Ctrl-C received")
    sys.exit(0)

def harvest_constants(options, flag, src, strconsts, errconsts,
                      nconsts, typedefs, apis, macros):
    global takeconst, takeapis, takedtypes, takemacros

    if flag:
        path = os.path.join(options.src, src)
    else:
        path = os.path.join(options.tgt, src)

    # open the file
    try:
        inputfile = open(path, "r", encoding="utf-8")
    except:
        path = os.path.join(options.includedir, src)
        try:
            inputfile = open(path, "r", encoding="utf-8")
        except:
            print("File", path, "could not be opened")
            return -1
    # read the file - these files aren't too large
    # so ingest the whole thing at one gulp
    lines = inputfile.readlines()
    # cache the string constants, numeric, and typedef constants
    # in separate lists
    strconstlen = 0
    nconstlen = 0
    # loop over the lines
    for n in range(len(lines)):
        line = lines[n]
        # remove white space at front and back
        myline = line.strip()
        # remove comment lines
        if "/*" in myline or "*/" in myline or myline.startswith("*"):
            n += 1
            continue
        # if the line starts with #define, then we want it
        if (takeconst or takemacros) and myline.startswith("#define"):
            value = myline[8:]
             # skip some well-known unwanted values
            if value.startswith("PMIx"):
                continue
            if value.startswith("PMIX_HAVE_VISIB"):
                continue
            if "__attribute__" in value:
                continue
            if takeconst:
                tokens = value.split()
                if len(tokens) >= 2:
                    if tokens[1][0] == '"':
                        strconsts.append([tokens[0], tokens[1]])
                        if len(tokens[0]) > strconstlen:
                            strconstlen = len(tokens[0])
                    elif "PMIX_ERR_" in value or tokens[1].startswith("-"):
                        # numerical constant that looks just like a
                        # string constant - i.e., PMIX_ERR_FOO...1
                        # we output them in a separate section, but
                        # consider them string constants
                        uniq = True
                        for a in errconsts:
                            if tokens[0] == a[0]:
                                uniq = False
                                break
                        if uniq:
                            errconsts.append([tokens[0], tokens[1]])
                            if len(tokens[0]) > strconstlen:
                                strconstlen = len(tokens[0])
                    elif tokens[1].isdigit() or tokens[1].startswith("UINT") or tokens[1].startswith("0x"):
                        uniq = True
                        for a in nconsts:
                            if tokens[0] == a[0]:
                                uniq = False
                                break
                        if uniq:
                            nconsts.append([tokens[0], tokens[1]])
                            if len(tokens[0]) > nconstlen:
                                nconstlen = len(tokens[0])
            if takemacros:
                tokens = value.split()
                if len(tokens) >= 2:
                    if value.startswith("PMIX_") and \
                        "//" not in value and (value[-1] == '\\' or value.endswith(")")):
                        # find the closing paren as we don't want the rest
                        idx = value.find(")") + 1
                        macros.append(value[:idx])
                        # if the line ends in continuation, then 
                        # we have to take lines until we hit the end of the macro
                        if value[-1] == '\\':
                            n += 1
                            value = lines[n].strip()
                            while n < len(lines) and value[-1] == '\\':
                                n += 1
                                value = lines[n].strip()

        elif takeapis and myline.startswith("PMIX_EXPORT"):
            value = myline[12:].strip()
            # this is the name of an API - these
            # are frequently multi-line, so collect
            # all of them
            if ";" in value:
                value = value[:-1]
                newapi = [value]
                uniq = True
                for a in apis:
                    if newapi == a:
                        uniq = False
                        break
                if uniq:
                    apis.append(newapi)
            else:
                newapi = [value]
                apirunning = True
                while apirunning:
                    n += 1
                    value = lines[n].strip()
                    if ";" in value:
                        apirunning = False
                        value = value[:-1]
                    newapi.append(value)
                uniq = True
                for a in apis:
                    if newapi == a:
                        uniq = False
                        break
                if uniq:
                    apis.append(newapi)
        elif takedtypes and myline.startswith("typedef"):
            # there are three types of typedef's in PMIx:
            #
            # 1. simple one-line typedef - e.g., "typedef int foo"
            #
            # 2. multi-line struct types
            #
            # 3. function definitions - by PMIx convention, these
            #    always have an "fn_t" in the name
            #
            # 4. enum definitions
            #
            # start with the fourth option by looking for "enum"
            if "enum" in myline:
                # each line after this one should contain a name
                # so we just assign a value sequentially.
                counter = 0
                n += 1
                value = lines[n].strip()
                if ',' in value:
                    value = value[:-1]
                while '}' not in value and n < len(lines):
                    tokens[0] = value
                    tokens[1] = str(counter)
                    counter += 1
                    nconsts.append([tokens[0], tokens[1]])
                    if len(tokens[0]) > nconstlen:
                        nconstlen = len(tokens[0])
                    n += 1
                    value = lines[n].strip()
                    if ',' in value:
                        value = value[:-1]
                # the termination line contains the type name
                # for this enum - declare it as integer here
                value = "typedef int " + value[2:]
                uniq = True
                for a in typedefs:
                    if value == a:
                        uniq = False
                        break
                if uniq:
                    typedefs.append([value])
            # now address the first option - detectable by
            # having a semi-colon at the end of the line and
            # no "fn_t" or "cbfunc_t" in it
            elif ";" in myline and not "fn_t" in myline and not "cbfunc_t" in myline:
                value = myline[:-1]
                # check for pre-declaration statements of form
                # typedef struct foo foo
                ck = value.split()
                if len(ck) == 4 and ck[1] == "struct" and ck[2] == ck[3]:
                    n += 1
                    continue
                else:
                    uniq = True
                    for a in typedefs:
                        if value == a:
                            uniq = False
                            break
                    if uniq:
                        typedefs.append([value])
            # now check the third option by looking for
            # "fn_t" or "cbfunc_t" in it
            elif "fn_t" in myline or "cbfunc_t" in myline:
                if ";" in myline:
                    # this is a one-line function definition
                    value = myline[:-1]
                    uniq = True
                    for a in typedefs:
                        if value == a:
                            uniq = False
                            break
                    if uniq:
                        typedefs.append([value])
                else:
                    # this is a multi-line function definition
                    newdef = [myline]
                    defrunning = True
                    while defrunning:
                        n += 1
                        value = lines[n].strip()
                        if ";" in value:
                            defrunning = False
                            value = value[:-1]
                        newdef.append(value)
                    if takeapis:
                        uniq = True
                        for a in apis:
                            if value == a:
                                uniq = False
                                break
                        if uniq:
                            apis.append(newdef)
            # must be a multi-line struct type definition
            # we capture these in the typedef list
            # so we output their type definition
            else:
                # check for pre-declaration statements of form
                # typedef struct foo foo
                value = myline
                ck = value.split()
                if len(ck) == 4 and ck[1] == "struct" and ck[2] == ck[3]:
                    n += 1
                    continue
                else:
                    newdef = []
                    n += 1
                    value = lines[n].strip()
                    nbrk = 1
                    while nbrk > 0:
                        # avoid comments
                        if "/*" in value or "*/" in value or value.startswith("*"):
                            n += 1
                            value = lines[n].strip()
                            continue
                        if "}" in value:
                            nbrk -= 1
                            if nbrk > 0:
                                n += 1
                                value = lines[n].strip()
                            continue
                        if  "union" in value:
                            # we have to create another definition that contains
                            # just the union and then add that back into the
                            # current typedef
                            uniondef = []
                            n += 1
                            value = lines[n].strip()
                            while "}" not in value:
                                # terminate at the semi-colon
                                idx = value.rfind(';')
                                value = "    " + value[:idx]
                                # add it to the union def
                                uniondef.append(value)
                                n += 1
                                value = lines[n].strip()
                            # terminate at the semi-colon
                            idx = value.rfind(';')
                            value = value[:idx]
                            # extract the name of the union
                            idx = value.rfind(' ')
                            value = value[idx+1:]
                            # save this name
                            lowname = value
                            # capitalize the first letter
                            value = value.title()
                            # create the union definition - the output
                            # routine will prepend a 'c'
                            myvalue = "def union " + value + ":"
                            uniondef.insert(0, myvalue)
                            typedefs.append(uniondef)
                            # add the union to the struct
                            value = "    " + value + " " + lowname
                            newdef.append(value)
                            nbrk -= 1
                            n += 1
                            value = lines[n].strip()
                            continue
                        # terminate at the semi-colon
                        idx = value.rfind(';')
                        value = "    " + value[:idx]
                        newdef.append(value)
                        n += 1
                        value = lines[n].strip()
                    # we need to extract the type name
                    value = "typedef struct " + value[2:]
                    value = value[:-1] + ":"
                    newdef.insert(0, value)
                    uniq = True
                    for a in typedefs:
                        if newdef == a:
                            uniq = False
                            break
                    if uniq:
                        typedefs.append(newdef)
    return 0

def main():
    global takeconst, takeapis, takedtypes, takemacros
    signal.signal(signal.SIGINT, signal_handler)

    parser = OptionParser("usage: %prog [options]")
    debugGroup = OptionGroup(parser, "Debug Options")
    debugGroup.add_option("--debug",
                          action="store_true", dest="debug", default=False,
                          help="Output lots of debug messages while processing")
    debugGroup.add_option("--dryrun",
                          action="store_true", dest="dryrun", default=False,
                          help="Show commands, but do not execute them")
    parser.add_option_group(debugGroup)

    execGroup = OptionGroup(parser, "Execution Options")
    execGroup.add_option("--src", dest="src",
                         help="The directory where the source PMIx header files will be found")
    execGroup.add_option("--tgt", dest="tgt",
                         help="The directory where the target PMIx header files will be found")
    execGroup.add_option("--constants",
                         action="store_true", dest="constants", default=False,
                         help="Check constants")
    execGroup.add_option("--apis",
                         action="store_true", dest="apis", default=False,
                         help="Check APIs")
    execGroup.add_option("--datatypes",
                         action="store_true", dest="datatypes", default=False,
                         help="Check datatypes")
    execGroup.add_option("--macros",
                         action="store_true", dest="macros", default=False,
                         help="Check macros")
    parser.add_option_group(execGroup)

    (options, args) = parser.parse_args()

    if not options.src:
        print("Must specify source directory")
        sys.exit(1)
    # see if the source directory exists
    if not os.path.exists(options.src):
        print("SOURCE directory",options.src,"does not exist")
        sys.exit(1)

    if not options.tgt:
        print("Must specify target directory")
        sys.exit(1)
    # see if the include directory directory exists
    if not os.path.exists(options.tgt):
        print("Target directory",options.includedir,"does not exist")
        sys.exit(1)

    if not options.constants and not options.apis and not options.datatypes and not options.macros:
        takeconst = True
        takeapis = True
        takedtypes = True
        takemacros = True

    if options.constants:
        takeconst = True
    if options.apis:
        takeapis = True
    if options.datatypes:
        takedtypes = True
    if options.macros:
        takemacros = True

    if options.dryrun or options.debug:
        debug = True
    else:
        debug = False

    # scan across the header files in the src directory
    # looking for constants and typedefs
    srcconsts = []
    srcerr = []
    srcnconsts = []
    srctypes = []
    srcapis = []
    srcmacros = []

    if harvest_constants(options, True, "pmix_common.h.in", srcconsts,
                         srcerr, srcnconsts, srctypes, srcapis, srcmacros) != 0:
        sys.exit(1)

    if harvest_constants(options, True, "pmix.h", srcconsts,
                         srcerr, srcnconsts, srctypes, srcapis, srcmacros) != 0:
        sys.exit(1)

    if harvest_constants(options, True, "pmix_server.h", srcconsts,
                         srcerr, srcnconsts, srctypes, srcapis, srcmacros) != 0:
        sys.exit(1)

    if harvest_constants(options, True, "pmix_tool.h", srcconsts,
                         srcerr, srcnconsts, srctypes, srcapis, srcmacros) != 0:
        sys.exit(1)

    if harvest_constants(options, True, "pmix_deprecated.h", srcconsts,
                         srcerr, srcnconsts, srctypes, srcapis, srcmacros) != 0:
        sys.exit(1)
    

    # do the same for the tgt directory
    tgtconsts = []
    tgterr = []
    tgtnconsts = []
    tgttypes = []
    tgtapis = []
    tgtmacros = []

    if harvest_constants(options, False, "pmix_common.h.in", tgtconsts,
                         tgterr, tgtnconsts, tgttypes, tgtapis, tgtmacros) != 0:
        sys.exit(1)

    if harvest_constants(options, False, "pmix.h", tgtconsts,
                         tgterr, tgtnconsts, tgttypes, tgtapis, tgtmacros) != 0:
        sys.exit(1)

    if harvest_constants(options, False, "pmix_server.h", tgtconsts,
                         tgterr, tgtnconsts, tgttypes, tgtapis, tgtmacros) != 0:
        sys.exit(1)

    if harvest_constants(options, False, "pmix_tool.h", tgtconsts,
                         tgterr, tgtnconsts, tgttypes, tgtapis, tgtmacros) != 0:
        sys.exit(1)

    if harvest_constants(options, False, "pmix_deprecated.h", tgtconsts,
                         tgterr, tgtnconsts, tgttypes, tgtapis, tgtmacros) != 0:
        sys.exit(1)


    added = []
    redefined = []
    removed = []

    # cross check the constants
    for s in srcconsts:
        match = False
        for t in tgtconsts:
            if s == t:
                match = True
                break
            elif s[0] == t[0]:
                match = True
                redefined.append(["Attributes", s])
                redefined.append(["Attributes", t])
                break;
        if not match:
            added.append(["Attributes", s])
    # check the reverse
    for t in tgtconsts:
        match = False
        for s in srcconsts:
            if s == t:
                match = True
                break
        if not match:
            removed.append(["Attributes", t])

    # cross check the error codes
    for s in srcerr:
        match = False
        for t in tgterr:
            if s == t:
                match = True
                break
            elif s[0] == t[0]:
                match = True
                redefined.append(["Error constants", s])
                redefined.append(["Error constants", t])
                break;
        if not match:
            added.append(["Error constants", s])
    # check the reverse
    for t in tgterr:
        match = False
        for s in srcerr:
            if s == t:
                match = True
                break
        if not match:
            removed.append(["Error constants", t])

    # cross check the numerical constants
    for s in srcnconsts:
        match = False
        for t in tgtnconsts:
            if s == t:
                match = True
                break
            elif s[0] == t[0]:
                match = True
                redefined.append(["Numerical constants", s])
                redefined.append(["Numerical constants", t])
                break;
        if not match:
            added.append(["Numerical constants", s])
    # check the reverse
    for t in tgtnconsts:
        match = False
        for s in srcnconsts:
            if s == t:
                match = True
                break
        if not match:
            removed.append(["Numerical constants", t])

    # cross check the data types
    for s in srctypes:
        match = False
        for t in tgttypes:
            if s == t:
                match = True
                break
            elif "pmix_server_module_t" in s[0] and "pmix_server_module_t" in t[0]:
                match = True
                # we have to check the server module fields individually
                submatch = False
                for sm in s[1:]:
                    smt = sm.split()
                    for tm in t[1:]:
                        tmt = tm.split()
                        if tmt == smt:
                            submatch = True
                            break
                    if not submatch:
                        redefined.append(["Server module functions", s[0]])
            elif s[0] == t[0]:
                match = True
                redefined.append(["Types", s])
                redefined.append(["Types", t])
                break;
        if not match:
            added.append(["Types", s])
    # check the reverse
    for t in tgttypes:
        # we already dealt with the server module special case
        if "pmix_server_module_t" in t[0]:
            continue
        match = False
        for s in srctypes:
            if s == t:
                match = True
                break
            elif "pmix_server_module_t" in s[0]:
                match = True
                continue
        if not match:
            removed.append(["Types", t])

    # cross check the APIs
    for s in srcapis:
        match = False
        for t in tgtapis:
            if s == t:
                match = True
                break
            elif s[0] == t[0]:
                match = True
                redefined.append(["APIs", s])
                redefined.append(["APIs", t])
                break;
        if not match:
            added.append(["APIs", s])
    # check the reverse
    for t in tgtapis:
        match = False
        for s in srcapis:
            if s == t:
                match = True
                break
        if not match:
            removed.append(["APIs", t])

    # cross check the macros
    for s in srcmacros:
        match = False
        for t in tgtmacros:
            if s == t:
                match = True
                break
        if not match:
            added.append(["Macros", s])
    # check the reverse
    for t in tgtmacros:
        match = False
        for s in srcmacros:
            if s == t:
                match = True
                break
        if not match:
            removed.append(["Macros", t])

    # output the results
    if 0 < len(added):
        print("The following items have been ADDED to the library")
        header = ""
        for i in added:
            if header != i[0]:
                print("\n", i[0].upper())
                header = i[0]
            if "APIS" == i[0].upper():
                loc = 0
                for a in i[1]:
                    if 0 == loc:
                        loc = a.find("(") + 1
                        if "fn_t" in a:
                            loc = a.find("(", loc) + 1
                        print("\t" + a)
                    else:
                        outstr = "\t" + a.rjust(len(a) + loc)
                        print(outstr)
            elif "MACROS" == i[0].upper():
                print("\t" + i[1][0])
            else:
                outstr = "\t" + ' = '.join(i[1])
                print(outstr)
    else:
        print("No items have been ADDED to the library")

    if 0 < len(redefined):
        print("\n\nThe following items have been MODIFIED in the library")
        header = ""
        i = 0
        while i < len(redefined):
            if header != redefined[i][0]:
                print("\n", redefined[i][0].upper())
                header = redefined[i][0]
            if "APIS" == redefined[i][0].upper():
                loc = 0
                for a in redefined[i][1]:
                    if 0 == loc:
                        loc = a.find("(") + 1
                        print("\t" + a)
                    else:
                        outstr = "\t" + a.rjust(len(a) + loc)
                        print(outstr)
                i += 1
                loc = 0
                for a in redefined[i][1]:
                    if 0 == loc:
                        loc = a.find("(") + 1
                        print("\t" + a)
                    else:
                        outstr = "\t" + a.rjust(len(a) + loc)
                        print(outstr)
                i += 1
            elif "MACROS" == redefined[i][0].upper():
                 print("\t" + redefined[i][1])
                 i += 1
                 print("\t" + redefined[i][1])
                 i += 1
            else:
                outstr = "\t" + ' = '.join(redefined[i][1])
                print(outstr)
                i += 1
                outstr = "\t" + ' = '.join(redefined[i][1])
                print(outstr)
                i += 1
    else:
        print("\n\nNo items have been MODIFIED in the library")

    if 0 < len(removed):
        print("\n\nThe following items have been REMOVED from the library")
        header = ""
        for i in removed:
            if header != i[0]:
                print("\n", i[0].upper())
                header = i[0]
            if "APIS" == header.upper():
                loc = 0
                for a in i[1]:
                    if 0 == loc:
                        loc = a.find("(") + 1
                        print("\t" + a)
                    else:
                        outstr = "\t" + a.rjust(len(a) + loc)
                        print(outstr)
            elif "MACROS" == i[0].upper():
                 print("\t" + i[1])
            else:
                outstr = "\t" + ' = '.join(i[1])
                print(outstr)
    else:
        print("\n\nNo items have been REMOVED from the library")

if __name__ == '__main__':
    main()




#!/usr/bin/env python3
#
# Copyright (c) 2025      Jeffrey M. Squyres.  All rights reserved.
# Copyright (c) 2025-2026 Nanook Consulting  All rights reserved.
# $COPYRIGHT$
#
# Additional copyrights may follow
#
# $HEADER$
#

from __future__ import print_function
import os
import sys
import argparse

def parse_cmd_line_options(path, verbose=False):
    # read the file and parse it for options
    options = []
    with open(path) as file:
        for line in file:
            stripped = line.strip()
            if stripped.startswith("#define"):
                # find the first space at end of "define"
                s1 = stripped.find(' ')
                # first word after "#define " is the name of the option
                start = 8
                end = stripped.find(' ', start+1)
                if -1 == end:
                    continue
                name = stripped[start:end]
                # find the first quote mark
                start = stripped.find('"', end)
                if start != -1:
                    # find the ending quote mark
                    end = stripped.find('"', start+1)
                    # everything in-between is the actual option
                    opt = stripped[start+1:end]
                    # track it
                    options.append((name, opt))

    return options

def find_files(root, verbose=False):
    # Check for existence of root directory - otherwise, we just
    # fall through with no error output
    if not os.path.isdir(root):
        sys.stderr.write("Root directory " + root + " does not exist\nCannot continue\n")
        exit(1)

    # Search for help-*.txt files across the source tree, skipping
    # some directories (e.g., .git)
    # Also find and retrieve all source files (.h or .c)
    # so we can search them for entries
    help_files = []
    source_files = []
    tool_help_files = []
    tool_source_files = []
    # skip infrastructure directories
    skip_dirs = ['.git', '.libs', '.deps']
    # there may also be some help files we want to ignore
    skip_files = []
    for root_dir, dirs, files in os.walk(root):
        for sd in skip_dirs:
            if sd in dirs:
                dirs.remove(sd)

        for file in files:
            if file.startswith("help-") and file.endswith(".txt"):
                skipit = False
                for sf in skip_files:
                    if sf == file:
                        skipit = True
                        break
                if not skipit:
                    full_path = os.path.join(root_dir, file)
                    if "tools" in full_path:
                        tool_help_files.append(full_path)
                        if verbose:
                            print("Found tool help: {full_path}".format(full_path=full_path))
                    else:
                        help_files.append(full_path)
                        if verbose:
                            print("Found help: {full_path}".format(full_path=full_path))
            elif file.endswith(".c") or file.endswith(".h"):
                full_path = os.path.join(root_dir, file)
                if "tools" in full_path:
                    tool_source_files.append(full_path)
                else:
                    source_files.append(full_path)
                if verbose:
                    print("Found source: {full_path}".format(full_path=full_path))

    return help_files, source_files, tool_help_files, tool_source_files

def parse_help_files(file_paths, data, citations, verbose=False):
    # Parse INI-style help files, returning a dictionary with filenames as
    # keys.  Don't use the Python configparse module in order to
    # reduce dependencies (i.e., so that we don't have to pip install
    # anything to run this script).
    for file_path in file_paths:
        sections = {}
        current_section = None
        with open(file_path) as file:
            for line in file:
                stripped = line.strip()
                if stripped.startswith("#include"):
                    # if we aren't in a section, then this is an error
                    if current_section is None:
                        sys.stderr.write("ERROR: INCLUDE is specified in ", file_path, " outside of a section")
                        continue
                    # include the line in the section
                    sections[current_section].append(stripped)
                    # this line includes a file/topic from another file
                    # find the '#' at the end of the "include" keyword
                    start = stripped.find('#', 2)
                    if start != -1:
                        # find the '#'' at the end of the filename
                        end = stripped.find('#', start+1)
                        end2 = stripped.find('#', end+1)
                        if end2 == -1:
                            topic = stripped[end+1:]
                            fil = stripped[start+1:end]
                            citations.append((fil, topic))
                        else:
                            topic = stripped[end2+1:]
                            fil = stripped[end+1:end2]
                            citations.append((fil, topic))
                    # if a section contains an include, we treat that
                    # section as having been cited
                    citations.append((os.path.basename(file_path), current_section))
                    continue
                if stripped.startswith('#'):
                    current_section = None
                    continue
                if current_section is None and not stripped:
                    continue
                if stripped.startswith('['):
                    # find the end of the section name
                    end = stripped.find(']', 1)
                    if -1 == end:
                        continue
                    current_section = stripped[1:end]
                    sections[current_section] = list()
                elif current_section is not None:
                    sections[current_section].append(stripped)

        if file_path in data:
            sys.stderr.write("ERROR: path " + file_path + " already exists in data dictionary\n")
        else:
            data[file_path] = sections
        if verbose:
            p = len(sections)
            print("Parsed: {file_path} ({p} sections found)".format(file_path=file_path, p=p))

    return

def parse_src_files(source_files, citations, verbose=False):
    # search the source code for show_help references
    for src in source_files:
        cont_topic = False
        cont_filename = False
        with open(src) as file:
            for line in file:
                line = line.strip()

                # skip the obvious comment lines
                if line.startswith("//") or line.startswith("/*") or line.startswith('*') or line.startswith("PMIX_EXPORT"):
                    continue

                if cont_topic:
                    # the topic for the prior filename should be on this line
                    if verbose:
                        print("CONT TOPIC LINE: ", line)
                    start = line.find('"')
                    if start != -1:
                        end = line.find('"', start+1, -1)
                        if end != -1:
                            topic = line[start+1:end]
                            citations.append((filename,topic))
                            if verbose:
                                print("Found topic: ", filename, topic)
                            cont_topic = False
                            continue
                elif cont_filename:
                    # the filename should be on this line
                    if verbose:
                        print("CONT FILENAME LINE: ", line)
                    start = line.find('"')
                    if start != -1:
                        end = line.find('"', start+1, -1)
                        if end != -1:
                            filename = line[start+1:end]
                            # see if the topic is on the same line
                            start = line.find('"', end+1, -1)
                            if start != -1:
                                # find the end of it
                                end = line.find('"', start+1, -1)
                                if end != -1:
                                    topic = line[start+1:end]
                                    if verbose:
                                        print("Found all: ", filename, topic)
                                    citations.append((filename,topic))
                            else:
                                # topic must be on next line
                                cont_topic = True
                                if verbose:
                                    print("Found filename: ", filename)
                            cont_filename = False
                            continue
                        else:
                            cont_filename = False
                            sys.stderr.write("ERROR: Missing end of filename\n")
                            continue

                if ("pmix_show_help(" in line and not "pmix_status_t" in line) or \
                   ("pmix_show_help_string(" in line and not "char *" in line) or \
                   ("send_error_show_help(" in line and not "void" in line):
                    cont_topic = False
                    cont_filename = False
                    # line contains call to show-help - try to extract
                    # filename/topic
                    if verbose:
                        print("LINE: ", line)
                    start = line.find('"')
                    if start != -1:
                        end = line.find('"', start+1, -1)
                        if end != -1:
                            filename = line[start+1:end]
                            # see if the topic is on this line
                            start = line.find('"', end+1, -1)
                            if start != -1:
                                # find the end of it
                                end = line.find('"', start+1, -1)
                                if end != -1:
                                    topic = line[start+1:end]
                                    if verbose:
                                        print("Found all: ", filename, topic)
                                    citations.append((filename,topic))
                            else:
                                # topic must be on next line
                                cont_topic = True
                                continue
                    else:
                        # the filename must be on next line
                        cont_filename = True
                        continue

    return

def parse_tool_files(help_files, source_files, cli_options, citations, verbose=False):
    # cycle through the tool help files
    for hlp in help_files:
        topics = []
        options = []
        cli = []
        local_cli = []
        insection = False
        # scan the help file to find cmd line options that
        # are referenced in the help file itself
        with open(hlp) as file:
            for line in file:
                line = line.strip()
                if not line:
                    continue
                if line.startswith('#'):
                    insection = False
                    continue
                if line.startswith('[') and line.endswith(']'):
                    # topic - save it
                    start = line.find('[') + 1
                    end = line.find(']')
                    topic = line[start:end]
                    topics.append(topic)
                    if topic != "usage":
                        insection = True
                    continue
                if not insection:
                    # scan the line for the "--" that indicates a cmd line option
                    opt = line.find("--")
                    if -1 == opt:
                        continue
                    # opt now points to the "--" in front of the option
                    opt += 2
                    optend = line.find(' ', opt+1)
                    option = line[opt:optend]
                    # track the option
                    options.append(option)
        # get the name of the tool this help file belongs to
        d = os.path.basename(hlp)
        # find the start of the tool name
        start = d.find('-') + 1
        end = d.find('.', start)
        tool = d[start:end]
        # search the source files for the corresponding tool
        # note that we need to look at .c AND .h files for definitions
        found = False
        for src in source_files:
            f = os.path.basename(src)
            end = f.find('.')
            tsrc = f[0:end]
            if tool == tsrc:
                found = True
                # search this source for cmd line options
                with open(src) as srcfile:
                    for lsrc in srcfile:
                        lsrc = lsrc.strip()
                        if lsrc.startswith("#define PMIX_CLI"):
                            # defining a local option
                            # find the first space at end of "define"
                            s1 = lsrc.find(' ')
                            # first word after "#define " is the name of the option
                            start = 8
                            end = lsrc.find(' ', start+1)
                            if -1 == end:
                                continue
                            name = lsrc[start:end]
                            # find the first quote mark
                            start = lsrc.find('"', end)
                            if start != -1:
                                # find the ending quote mark
                                end = lsrc.find('"', start+1)
                                # everything in-between is the actual option
                                opt = lsrc[start+1:end]
                                # track it
                                local_cli.append((name, opt))

                        elif lsrc.startswith("PMIX_OPTION_") and lsrc != "PMIX_OPTION_END":
                            # the option name starts with PMIX_CLI
                            start = lsrc.find("PMIX_CLI_")
                            end = lsrc.find(",", start+1)
                            # find the option in the global list so we can
                            # compare it to the help file options
                            cli.append(lsrc[start:end])

        if not found:
            sys.stderr.write("WARNING: No corresponding source code found for help file " + hlp + "\n")
            exit(1)

        # convert the CLI to strings so we can check the options and topics
        clistrings = []
        for c in cli:
            found = False
            # check local definitions first
            for l in local_cli:
                if c == l[0]:
                    found = True
                    clistrings.append(l[1])
                    break
            if not found:
                # check the global definitions
                for copt in cli_options:
                    if c == copt[0]:
                        found = True
                        clistrings.append(copt[1])
                        break
            if not found:
                # unrecognized option
                sys.stderr.write("WARNING: Unrecognized command line option " + c + " in source code " + os.path.basename(src) + "\n")
                exit(1)

        # check the cli and options to ensure they match
        for option in options:
            found = False
            for c in clistrings:
                # there must be a CLI entry for every option
                if c == option:
                    found = True
                    break
            if not found:
                sys.stderr.write("WARNING: Option " + option + " has no corresponding CLI defined for " + tool + "\n")
                exit(1)

        # reverse must also be true
        for c in clistrings:
            # ignore "help", "version", and "verbose" as those are handled
            # by the PMIx cmd line processor
            if "help" == c or "version" == c or "verbose" == c:
                continue
            found = False
            for option in options:
                # must be an option in the help file for each CLI entry
                if c == option:
                    found = True
                    break
            if not found:
                sys.stderr.write("WARNING: CLI definition " + c + " has no corresponding help entry in " + os.path.basename(hlp) + "\n")
                exit(1)

        # also require that there be a topic for each option so that
        # "--help option" works. Note that "help", "version", and "verbose"
        # are special cases handled by the cmd line processor
        # scan the options
        for option in options:
            if option == "help" or option == "version" or option == "verbose":
                continue
            # scan the topics for a match
            found = False
            for topic in topics:
                if option == topic:
                    if verbose:
                        print("CITED ", os.path.basename(hlp), option)
                    found = True
                    citations.append((os.path.basename(hlp), option))
                    break
            if not found:
                sys.stderr.write("WARNING: Option " + option + " has no topic entry in " + os.path.basename(hlp) + "\n")


def purge(parsed_data, citations):
    result_data = {}
    errorFound = False
    for filename in parsed_data:
        # tools were processed separately
        sections = parsed_data[filename]
        if "tools" in filename:
            result_data[filename] = sections
            continue
        result_sections = {}
        for section in sections:
            content_list = sections[section]
            # check for duplicate entries
            content = '\n'.join(content_list)
            # search all other entries for a matching section
            for (file2, sec) in parsed_data.items():
                if file2 == filename:
                    continue
                for (sec2, cl) in sec.items():
                    cnt = '\n'.join(cl)
                    if sec == section:
                        if content == cnt:
                            # these are the same
                            sys.stderr.write("DUPLICATE FOUND - SECTION: " + section + "\nFILES: " + filename + "\n       " + file2 + "\n")
                            errorFound = True
                        else:
                            # same topic, different content
                            sys.stderr.write("DUPLICATE SECTION WITH DIFFERENT CONTENT: " + section, "\nFILES: " + filename + "\n       " + file2 + "\n")
                            errorFound = True
            # search code files for usage
            # protect special values
            if section == "help" or section == "version" or section == "usage":
                result_sections[section] = content_list
                continue
            for entry in citations:
                used = False
                file2 = entry[0]
                sec = entry[1]
                if file2 == os.path.basename(filename) and sec == section:
                    # this is a used entry
                    result_sections[section] = content_list
                    used = True
                    break;
            if not used:
                sys.stderr.write("** WARNING: Unused help topic\n")
                sys.stderr.write("    File: " + filename + "\n")
                sys.stderr.write("    Section: " + section + "\n")
                errorFound = True
        # see if anything is left
        if 0 < len(result_sections):
            # don't retain the file if no citations for it are left
            result_data[filename] = result_sections
        else:
            sys.stderr.write("File " + filename + "has no used topics - omitting\n")
            errorFound = True

    if errorFound:
        exit(1)
    return result_data

def generate_c_code(parsed_data):
    # Generate C code with an array of filenames and their
    # corresponding INI sections.
    c_code = "// THIS FILE IS GENERATED AUTOMATICALLY! EDITS WILL BE LOST!\n// This file generated by " + os.path.basename(sys.argv[0]) + "\n\n"

    c_code += "#include \"src/include/pmix_config.h\"\n#include <stdio.h>\n#include <string.h>\n\n#include \"src/include/pmix_globals.h\"\n\n"

    ini_arrays = []
    file_entries = []

    for idx, (fil, sections) in enumerate(sorted(parsed_data.items())):
        filename = os.path.basename(fil)
        var_name = filename.replace('-', '_').replace('.', '_')

        ini_entries = []
        for section, content_list in sections.items():
            entry = "    { .topic = \"" + section + "\",\n      .content = (const char *[]){\n"
            for content in content_list:
                c_content = content.replace('"','\\"').replace("\n", '\\n"\n"').replace("\\;", ";")
                entry += "                 \"" + c_content + "\",\n"
            entry += "                 NULL}\n    },\n"
            ini_entries.append(entry)
        ini_entries.append("    { .topic = NULL, .content = NULL }")

        ini_array_name = "ini_entries_" + str(idx)
        ini_arrays.append("static pmix_show_help_entry_t " + ini_array_name + "[] = {\n" + "\n".join(ini_entries) + "\n};\n")
        file_entries.append("    { \"" + filename + "\", " + ini_array_name + "}")
    file_entries.append("    { NULL, NULL }")

    c_code += "\n".join(ini_arrays) + "\n"
    c_code += "pmix_show_help_file_t pmix_show_help_data[] = {\n" + ",\n".join(file_entries) + "\n};\n"

    return c_code

#-------------------------------

def main():
    parser = argparse.ArgumentParser(description="Generate C code from help text INI files.")
    parser.add_argument("--root",
                        required=True,
                        help="Root directory to search for help-*.txt files")
    parser.add_argument("--out",
                        required=True,
                        help="Output C file")
    parser.add_argument("--verbose",
                        action="store_true",
                        help="Enable verbose output")
    parser.add_argument("--purge",
                        action="store_true",
                        help="Purge duplicates, update topic pointers as required")
    parser.add_argument("--dryrun",
                        action="store_true",
                        help="Do not write out resulting ini_array_name")

    args = parser.parse_args()

    parsed_data = {}
    citations = []
    cli_options = []
    outdata = {}
    rootdir = os.path.join(args.root, "src")

    # the options are in root/util/pmix_cmd_line.h
    path = os.path.join(rootdir, "util", "pmix_cmd_line.h")
    if not os.path.exists(path):
        sys.stderr.write("File " + path + " does not exist\nCannot continue\n")
        exit(1)
    # obtain a list of (option name, string) tuples
    cli_options = parse_cmd_line_options(path, args.verbose)

    help_files, source_files, tool_help_files, tool_source_files = find_files(rootdir, args.verbose)
    parse_help_files(help_files, parsed_data, citations, args.verbose)
    parse_help_files(tool_help_files, parsed_data, citations, args.verbose)
    parse_src_files(source_files, citations, args.verbose)
    parse_tool_files(tool_help_files, tool_source_files, cli_options, citations, args.verbose)
    if args.purge:
        outdata = purge(parsed_data, citations)
    else:
        outdata = parsed_data

    if not args.dryrun:
        c_code = generate_c_code(outdata)

        with open(args.out, "w") as f:
            f.write(c_code)

        if args.verbose:
            print("Generated C code written to ", args.out)

if __name__ == "__main__":
    main()
